#    Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime

from octavia_lib.api.drivers import data_models as driver_dm
from octavia_lib.common import constants as lib_consts
from oslo_utils import uuidutils

from octavia.common import constants
from octavia.common import data_models


class SampleDriverDataModels(object):

    def __init__(self):
        self.project_id = uuidutils.generate_uuid()
        self.lb_id = uuidutils.generate_uuid()
        self.ip_address = '192.0.2.30'
        self.port_id = uuidutils.generate_uuid()
        self.network_id = uuidutils.generate_uuid()
        self.subnet_id = uuidutils.generate_uuid()
        self.qos_policy_id = uuidutils.generate_uuid()
        self.lb_name = uuidutils.generate_uuid()
        self.lb_description = uuidutils.generate_uuid()
        self.flavor_id = uuidutils.generate_uuid()
        self.flavor_profile_id = uuidutils.generate_uuid()

        self.listener1_id = uuidutils.generate_uuid()
        self.listener2_id = uuidutils.generate_uuid()
        self.default_tls_container_ref = uuidutils.generate_uuid()
        self.sni_container_ref_1 = uuidutils.generate_uuid()
        self.sni_container_ref_2 = uuidutils.generate_uuid()
        self.client_ca_tls_certificate_ref = uuidutils.generate_uuid()
        self.client_crl_container_ref = uuidutils.generate_uuid()
        self.pool_sni_container_ref = uuidutils.generate_uuid()
        self.pool_ca_container_ref = uuidutils.generate_uuid()
        self.pool_crl_container_ref = uuidutils.generate_uuid()

        self.pool1_id = uuidutils.generate_uuid()
        self.pool2_id = uuidutils.generate_uuid()

        self.hm1_id = uuidutils.generate_uuid()
        self.hm2_id = uuidutils.generate_uuid()

        self.member1_id = uuidutils.generate_uuid()
        self.member2_id = uuidutils.generate_uuid()
        self.member3_id = uuidutils.generate_uuid()
        self.member4_id = uuidutils.generate_uuid()

        self.l7policy1_id = uuidutils.generate_uuid()
        self.l7policy2_id = uuidutils.generate_uuid()

        self.l7rule1_id = uuidutils.generate_uuid()
        self.l7rule2_id = uuidutils.generate_uuid()

        self.created_at = datetime.datetime.now()
        self.updated_at = (datetime.datetime.now() +
                           datetime.timedelta(minutes=1))

        self._common_test_dict = {
            lib_consts.PROVISIONING_STATUS: constants.ACTIVE,
            lib_consts.OPERATING_STATUS: constants.ONLINE,
            lib_consts.PROJECT_ID: self.project_id,
            constants.CREATED_AT: self.created_at,
            constants.UPDATED_AT: self.updated_at,
            constants.ENABLED: True}

        # Setup Health Monitors
        self.test_hm1_dict = {
            lib_consts.ID: self.hm1_id,
            lib_consts.TYPE: constants.HEALTH_MONITOR_PING,
            lib_consts.DELAY: 1, lib_consts.TIMEOUT: 3,
            lib_consts.FALL_THRESHOLD: 1, lib_consts.RISE_THRESHOLD: 2,
            lib_consts.HTTP_METHOD: lib_consts.HEALTH_MONITOR_HTTP_METHOD_GET,
            lib_consts.URL_PATH: '/', lib_consts.EXPECTED_CODES: '200',
            lib_consts.NAME: 'hm1', lib_consts.POOL_ID: self.pool1_id,
            lib_consts.HTTP_VERSION: 1.0, lib_consts.DOMAIN_NAME: None,
            lib_consts.PROJECT_ID: self.project_id}

        self.test_hm1_dict.update(self._common_test_dict)

        self.test_hm2_dict = copy.deepcopy(self.test_hm1_dict)
        self.test_hm2_dict[lib_consts.ID] = self.hm2_id
        self.test_hm2_dict[lib_consts.NAME] = 'hm2'
        self.test_hm2_dict.update(
            {lib_consts.HTTP_VERSION: 1.1,
             lib_consts.DOMAIN_NAME: 'testdomainname.com'})

        self.db_hm1 = data_models.HealthMonitor(**self.test_hm1_dict)
        self.db_hm2 = data_models.HealthMonitor(**self.test_hm2_dict)

        self.provider_hm1_dict = {
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.DELAY: 1, lib_consts.EXPECTED_CODES: '200',
            lib_consts.HEALTHMONITOR_ID: self.hm1_id,
            lib_consts.HTTP_METHOD: lib_consts.HEALTH_MONITOR_HTTP_METHOD_GET,
            lib_consts.MAX_RETRIES: 2,
            lib_consts.MAX_RETRIES_DOWN: 1,
            lib_consts.NAME: 'hm1',
            lib_consts.POOL_ID: self.pool1_id,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.TIMEOUT: 3,
            lib_consts.TYPE: constants.HEALTH_MONITOR_PING,
            lib_consts.URL_PATH: '/',
            lib_consts.HTTP_VERSION: 1.0,
            lib_consts.DOMAIN_NAME: None}

        self.provider_hm2_dict = copy.deepcopy(self.provider_hm1_dict)
        self.provider_hm2_dict[lib_consts.HEALTHMONITOR_ID] = self.hm2_id
        self.provider_hm2_dict[lib_consts.NAME] = 'hm2'
        self.provider_hm2_dict.update(
            {lib_consts.HTTP_VERSION: 1.1,
             lib_consts.DOMAIN_NAME: 'testdomainname.com'})

        self.provider_hm1 = driver_dm.HealthMonitor(**self.provider_hm1_dict)
        self.provider_hm2 = driver_dm.HealthMonitor(**self.provider_hm2_dict)

        # Setup Members
        self.test_member1_dict = {
            lib_consts.ID: self.member1_id,
            lib_consts.POOL_ID: self.pool1_id,
            constants.IP_ADDRESS: '192.0.2.16',
            lib_consts.PROTOCOL_PORT: 80, lib_consts.WEIGHT: 0,
            lib_consts.BACKUP: False,
            lib_consts.SUBNET_ID: self.subnet_id,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.NAME: 'member1',
            lib_consts.OPERATING_STATUS: lib_consts.ONLINE,
            lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
            constants.ENABLED: True,
            constants.CREATED_AT: self.created_at,
            constants.UPDATED_AT: self.updated_at,
            lib_consts.MONITOR_ADDRESS: '192.0.2.26',
            lib_consts.MONITOR_PORT: 81}

        self.test_member1_dict.update(self._common_test_dict)

        self.test_member2_dict = copy.deepcopy(self.test_member1_dict)
        self.test_member2_dict[lib_consts.ID] = self.member2_id
        self.test_member2_dict[constants.IP_ADDRESS] = '192.0.2.17'
        self.test_member2_dict[lib_consts.MONITOR_ADDRESS] = '192.0.2.27'
        self.test_member2_dict[lib_consts.NAME] = 'member2'

        self.test_member3_dict = copy.deepcopy(self.test_member1_dict)
        self.test_member3_dict[lib_consts.ID] = self.member3_id
        self.test_member3_dict[constants.IP_ADDRESS] = '192.0.2.18'
        self.test_member3_dict[lib_consts.MONITOR_ADDRESS] = '192.0.2.28'
        self.test_member3_dict[lib_consts.NAME] = 'member3'
        self.test_member3_dict[lib_consts.POOL_ID] = self.pool2_id

        self.test_member4_dict = copy.deepcopy(self.test_member1_dict)
        self.test_member4_dict[lib_consts.ID] = self.member4_id
        self.test_member4_dict[constants.IP_ADDRESS] = '192.0.2.19'
        self.test_member4_dict[lib_consts.MONITOR_ADDRESS] = '192.0.2.29'
        self.test_member4_dict[lib_consts.NAME] = 'member4'
        self.test_member4_dict[lib_consts.POOL_ID] = self.pool2_id

        self.test_pool1_members_dict = [self.test_member1_dict,
                                        self.test_member2_dict]
        self.test_pool2_members_dict = [self.test_member3_dict,
                                        self.test_member4_dict]

        self.db_member1 = data_models.Member(**self.test_member1_dict)
        self.db_member2 = data_models.Member(**self.test_member2_dict)
        self.db_member3 = data_models.Member(**self.test_member3_dict)
        self.db_member4 = data_models.Member(**self.test_member4_dict)

        self.db_pool1_members = [self.db_member1, self.db_member2]
        self.db_pool2_members = [self.db_member3, self.db_member4]

        self.provider_member1_dict = {lib_consts.ADDRESS: '192.0.2.16',
                                      lib_consts.ADMIN_STATE_UP: True,
                                      lib_consts.MEMBER_ID: self.member1_id,
                                      lib_consts.MONITOR_ADDRESS: '192.0.2.26',
                                      lib_consts.MONITOR_PORT: 81,
                                      lib_consts.NAME: 'member1',
                                      lib_consts.POOL_ID: self.pool1_id,
                                      lib_consts.PROJECT_ID: self.project_id,
                                      lib_consts.PROTOCOL_PORT: 80,
                                      lib_consts.SUBNET_ID: self.subnet_id,
                                      lib_consts.WEIGHT: 0,
                                      lib_consts.BACKUP: False}

        self.provider_member2_dict = copy.deepcopy(self.provider_member1_dict)
        self.provider_member2_dict[lib_consts.MEMBER_ID] = self.member2_id
        self.provider_member2_dict[lib_consts.ADDRESS] = '192.0.2.17'
        self.provider_member2_dict[lib_consts.MONITOR_ADDRESS] = '192.0.2.27'
        self.provider_member2_dict[lib_consts.NAME] = 'member2'

        self.provider_member3_dict = copy.deepcopy(self.provider_member1_dict)
        self.provider_member3_dict[lib_consts.MEMBER_ID] = self.member3_id
        self.provider_member3_dict[lib_consts.ADDRESS] = '192.0.2.18'
        self.provider_member3_dict[lib_consts.MONITOR_ADDRESS] = '192.0.2.28'
        self.provider_member3_dict[lib_consts.NAME] = 'member3'
        self.provider_member3_dict[lib_consts.POOL_ID] = self.pool2_id

        self.provider_member4_dict = copy.deepcopy(self.provider_member1_dict)
        self.provider_member4_dict[lib_consts.MEMBER_ID] = self.member4_id
        self.provider_member4_dict[lib_consts.ADDRESS] = '192.0.2.19'
        self.provider_member4_dict[lib_consts.MONITOR_ADDRESS] = '192.0.2.29'
        self.provider_member4_dict[lib_consts.NAME] = 'member4'
        self.provider_member4_dict[lib_consts.POOL_ID] = self.pool2_id

        self.provider_pool1_members_dict = [self.provider_member1_dict,
                                            self.provider_member2_dict]

        self.provider_pool2_members_dict = [self.provider_member3_dict,
                                            self.provider_member4_dict]

        self.provider_member1 = driver_dm.Member(**self.provider_member1_dict)
        self.provider_member2 = driver_dm.Member(**self.provider_member2_dict)
        self.provider_member3 = driver_dm.Member(**self.provider_member3_dict)
        self.provider_member4 = driver_dm.Member(**self.provider_member4_dict)

        self.provider_pool1_members = [self.provider_member1,
                                       self.provider_member2]
        self.provider_pool2_members = [self.provider_member3,
                                       self.provider_member4]

        # Setup test pools
        self.test_pool1_dict = {
            lib_consts.ID: self.pool1_id,
            lib_consts.NAME: 'pool1', lib_consts.DESCRIPTION: 'Pool 1',
            constants.LOAD_BALANCER_ID: self.lb_id,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
            lib_consts.LB_ALGORITHM: lib_consts.LB_ALGORITHM_ROUND_ROBIN,
            lib_consts.MEMBERS: self.test_pool1_members_dict,
            constants.HEALTH_MONITOR: self.test_hm1_dict,
            lib_consts.SESSION_PERSISTENCE: {
                lib_consts.TYPE: lib_consts.LB_ALGORITHM_SOURCE_IP},
            lib_consts.LISTENERS: [],
            lib_consts.L7POLICIES: [],
            constants.TLS_CERTIFICATE_ID: self.pool_sni_container_ref,
            constants.CA_TLS_CERTIFICATE_ID: self.pool_ca_container_ref,
            constants.CRL_CONTAINER_ID: self.pool_crl_container_ref,
            lib_consts.TLS_ENABLED: True,
            lib_consts.TLS_CIPHERS: None,
            lib_consts.TLS_VERSIONS: None,
            lib_consts.ALPN_PROTOCOLS:
                constants.AMPHORA_SUPPORTED_ALPN_PROTOCOLS
        }

        self.test_pool1_dict.update(self._common_test_dict)

        self.test_pool2_dict = copy.deepcopy(self.test_pool1_dict)
        self.test_pool2_dict[lib_consts.ID] = self.pool2_id
        self.test_pool2_dict[lib_consts.NAME] = 'pool2'
        self.test_pool2_dict[lib_consts.DESCRIPTION] = 'Pool 2'
        self.test_pool2_dict[
            lib_consts.MEMBERS] = self.test_pool2_members_dict
        del self.test_pool2_dict[constants.TLS_CERTIFICATE_ID]
        del self.test_pool2_dict[constants.CA_TLS_CERTIFICATE_ID]
        del self.test_pool2_dict[constants.CRL_CONTAINER_ID]

        self.test_pools = [self.test_pool1_dict, self.test_pool2_dict]

        self.db_pool1 = data_models.Pool(**self.test_pool1_dict)
        self.db_pool1.health_monitor = self.db_hm1
        self.db_pool1.members = self.db_pool1_members
        self.db_pool2 = data_models.Pool(**self.test_pool2_dict)
        self.db_pool2.health_monitor = self.db_hm2
        self.db_pool2.members = self.db_pool2_members

        self.test_db_pools = [self.db_pool1, self.db_pool2]
        pool_cert = data_models.TLSContainer(certificate='pool cert')
        pool_ca_file_content = 'X509 POOL CA CERT FILE'
        pool_crl_file_content = 'X509 POOL CRL FILE'

        self.provider_pool1_dict = {
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.DESCRIPTION: 'Pool 1',
            lib_consts.HEALTHMONITOR: self.provider_hm1_dict,
            lib_consts.LB_ALGORITHM: lib_consts.LB_ALGORITHM_ROUND_ROBIN,
            lib_consts.LOADBALANCER_ID: self.lb_id,
            lib_consts.MEMBERS: self.provider_pool1_members_dict,
            lib_consts.NAME: 'pool1',
            lib_consts.POOL_ID: self.pool1_id,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
            lib_consts.SESSION_PERSISTENCE: {
                lib_consts.TYPE: lib_consts.LB_ALGORITHM_SOURCE_IP},
            lib_consts.TLS_CONTAINER_REF: self.pool_sni_container_ref,
            lib_consts.TLS_CONTAINER_DATA: pool_cert.to_dict(),
            lib_consts.CA_TLS_CONTAINER_REF: self.pool_ca_container_ref,
            lib_consts.CA_TLS_CONTAINER_DATA: pool_ca_file_content,
            lib_consts.CRL_CONTAINER_REF: self.pool_crl_container_ref,
            lib_consts.CRL_CONTAINER_DATA: pool_crl_file_content,
            lib_consts.TLS_ENABLED: True,
            lib_consts.TLS_CIPHERS: None,
            lib_consts.TLS_VERSIONS: None,
            lib_consts.ALPN_PROTOCOLS:
                constants.AMPHORA_SUPPORTED_ALPN_PROTOCOLS
        }

        self.provider_pool2_dict = copy.deepcopy(self.provider_pool1_dict)
        self.provider_pool2_dict[lib_consts.POOL_ID] = self.pool2_id
        self.provider_pool2_dict[lib_consts.NAME] = 'pool2'
        self.provider_pool2_dict[lib_consts.DESCRIPTION] = 'Pool 2'
        self.provider_pool2_dict[
            lib_consts.MEMBERS] = self.provider_pool2_members_dict
        self.provider_pool2_dict[
            lib_consts.HEALTHMONITOR] = self.provider_hm2_dict
        self.provider_pool2_dict[lib_consts.TLS_CONTAINER_REF] = None
        del self.provider_pool2_dict[lib_consts.TLS_CONTAINER_DATA]
        self.provider_pool2_dict[lib_consts.CA_TLS_CONTAINER_REF] = None
        del self.provider_pool2_dict[lib_consts.CA_TLS_CONTAINER_DATA]
        self.provider_pool2_dict[lib_consts.CRL_CONTAINER_REF] = None
        del self.provider_pool2_dict[lib_consts.CRL_CONTAINER_DATA]

        self.provider_pool1 = driver_dm.Pool(**self.provider_pool1_dict)
        self.provider_pool1.members = self.provider_pool1_members
        self.provider_pool1.healthmonitor = self.provider_hm1
        self.provider_pool2 = driver_dm.Pool(**self.provider_pool2_dict)
        self.provider_pool2.members = self.provider_pool2_members
        self.provider_pool2.healthmonitor = self.provider_hm2

        self.provider_pools = [self.provider_pool1, self.provider_pool2]

        # Setup L7Rules
        self.test_l7rule1_dict = {
            lib_consts.ID: self.l7rule1_id,
            lib_consts.L7POLICY_ID: self.l7policy1_id,
            lib_consts.TYPE: lib_consts.L7RULE_TYPE_PATH,
            lib_consts.COMPARE_TYPE: lib_consts.L7RULE_COMPARE_TYPE_EQUAL_TO,
            lib_consts.KEY: 'fake_key',
            lib_consts.VALUE: 'fake_value',
            lib_consts.PROJECT_ID: self.project_id,
            constants.L7POLICY: None,
            lib_consts.INVERT: False}

        self.test_l7rule1_dict.update(self._common_test_dict)

        self.test_l7rule2_dict = copy.deepcopy(self.test_l7rule1_dict)
        self.test_l7rule2_dict[lib_consts.ID] = self.l7rule2_id

        self.test_l7rules = [self.test_l7rule1_dict, self.test_l7rule2_dict]

        self.db_l7Rule1 = data_models.L7Rule(**self.test_l7rule1_dict)
        self.db_l7Rule2 = data_models.L7Rule(**self.test_l7rule2_dict)

        self.db_l7Rules = [self.db_l7Rule1, self.db_l7Rule2]

        self.provider_l7rule1_dict = {
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.COMPARE_TYPE: lib_consts.L7RULE_COMPARE_TYPE_EQUAL_TO,
            lib_consts.INVERT: False,
            lib_consts.KEY: 'fake_key',
            lib_consts.L7POLICY_ID: self.l7policy1_id,
            lib_consts.L7RULE_ID: self.l7rule1_id,
            lib_consts.TYPE: lib_consts.L7RULE_TYPE_PATH,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.VALUE: 'fake_value'}

        self.provider_l7rule2_dict = copy.deepcopy(self.provider_l7rule1_dict)
        self.provider_l7rule2_dict[lib_consts.L7RULE_ID] = self.l7rule2_id
        self.provider_l7rules_dicts = [self.provider_l7rule1_dict,
                                       self.provider_l7rule2_dict]

        self.provider_l7rule1 = driver_dm.L7Rule(**self.provider_l7rule1_dict)
        self.provider_l7rule2 = driver_dm.L7Rule(**self.provider_l7rule2_dict)

        self.provider_rules = [self.provider_l7rule1, self.provider_l7rule2]

        # Setup L7Policies
        self.test_l7policy1_dict = {
            lib_consts.ID: self.l7policy1_id,
            lib_consts.NAME: 'l7policy_1',
            lib_consts.DESCRIPTION: 'L7policy 1',
            lib_consts.LISTENER_ID: self.listener1_id,
            lib_consts.ACTION: lib_consts.L7POLICY_ACTION_REDIRECT_TO_URL,
            lib_consts.REDIRECT_POOL_ID: None,
            lib_consts.REDIRECT_URL: 'http://example.com/index.html',
            lib_consts.REDIRECT_PREFIX: None,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.POSITION: 1,
            constants.LISTENER: None,
            constants.REDIRECT_POOL: None,
            lib_consts.L7RULES: self.test_l7rules,
            lib_consts.REDIRECT_HTTP_CODE: 302}

        self.test_l7policy1_dict.update(self._common_test_dict)

        self.test_l7policy2_dict = copy.deepcopy(self.test_l7policy1_dict)
        self.test_l7policy2_dict[lib_consts.ID] = self.l7policy2_id
        self.test_l7policy2_dict[lib_consts.NAME] = 'l7policy_2'
        self.test_l7policy2_dict[lib_consts.DESCRIPTION] = 'L7policy 2'

        self.test_l7policies = [self.test_l7policy1_dict,
                                self.test_l7policy2_dict]

        self.db_l7policy1 = data_models.L7Policy(**self.test_l7policy1_dict)
        self.db_l7policy2 = data_models.L7Policy(**self.test_l7policy2_dict)
        self.db_l7policy1.l7rules = self.db_l7Rules
        self.db_l7policy2.l7rules = self.db_l7Rules

        self.db_l7policies = [self.db_l7policy1, self.db_l7policy2]

        self.provider_l7policy1_dict = {
            lib_consts.ACTION: lib_consts.L7POLICY_ACTION_REDIRECT_TO_URL,
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.DESCRIPTION: 'L7policy 1',
            lib_consts.L7POLICY_ID: self.l7policy1_id,
            lib_consts.LISTENER_ID: self.listener1_id,
            lib_consts.NAME: 'l7policy_1',
            lib_consts.POSITION: 1,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.REDIRECT_POOL_ID: None,
            lib_consts.REDIRECT_URL: 'http://example.com/index.html',
            lib_consts.REDIRECT_PREFIX: None,
            lib_consts.RULES: self.provider_l7rules_dicts,
            lib_consts.REDIRECT_HTTP_CODE: 302
        }

        self.provider_l7policy2_dict = copy.deepcopy(
            self.provider_l7policy1_dict)
        self.provider_l7policy2_dict[
            lib_consts.L7POLICY_ID] = self.l7policy2_id
        self.provider_l7policy2_dict[lib_consts.NAME] = 'l7policy_2'
        self.provider_l7policy2_dict[lib_consts.DESCRIPTION] = 'L7policy 2'

        self.provider_l7policies_dict = [self.provider_l7policy1_dict,
                                         self.provider_l7policy2_dict]

        self.provider_l7policy1 = driver_dm.L7Policy(
            **self.provider_l7policy1_dict)
        self.provider_l7policy1.rules = self.provider_rules
        self.provider_l7policy2 = driver_dm.L7Policy(
            **self.provider_l7policy2_dict)
        self.provider_l7policy2.rules = self.provider_rules

        self.provider_l7policies = [self.provider_l7policy1,
                                    self.provider_l7policy2]

        # Setup Listeners
        self.test_listener1_dict = {
            lib_consts.ID: self.listener1_id,
            lib_consts.NAME: 'listener_1',
            lib_consts.DESCRIPTION: 'Listener 1',
            lib_consts.DEFAULT_POOL_ID: self.pool1_id,
            constants.LOAD_BALANCER_ID: self.lb_id,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
            lib_consts.PROTOCOL_PORT: 90,
            lib_consts.CONNECTION_LIMIT: 10000,
            constants.TLS_CERTIFICATE_ID: self.default_tls_container_ref,
            lib_consts.DEFAULT_POOL: self.test_pool1_dict,
            constants.SNI_CONTAINERS: [self.sni_container_ref_1,
                                       self.sni_container_ref_2],
            constants.PEER_PORT: 55,
            lib_consts.L7POLICIES: self.test_l7policies,
            lib_consts.INSERT_HEADERS: {},
            lib_consts.TIMEOUT_CLIENT_DATA: 1000,
            lib_consts.TIMEOUT_MEMBER_CONNECT: 2000,
            lib_consts.TIMEOUT_MEMBER_DATA: 3000,
            lib_consts.TIMEOUT_TCP_INSPECT: 4000,
            constants.CLIENT_CA_TLS_CERTIFICATE_ID:
                self.client_ca_tls_certificate_ref,
            lib_consts.CLIENT_AUTHENTICATION: constants.CLIENT_AUTH_NONE,
            constants.CLIENT_CRL_CONTAINER_ID: self.client_crl_container_ref,
            lib_consts.ALLOWED_CIDRS: ['192.0.2.0/24', '198.51.100.0/24'],
            lib_consts.TLS_CIPHERS: constants.CIPHERS_OWASP_SUITE_B,
            lib_consts.TLS_VERSIONS: constants.TLS_VERSIONS_OWASP_SUITE_B,
            lib_consts.ALPN_PROTOCOLS:
                constants.AMPHORA_SUPPORTED_ALPN_PROTOCOLS
        }

        self.test_listener1_dict.update(self._common_test_dict)

        self.test_listener2_dict = copy.deepcopy(self.test_listener1_dict)
        self.test_listener2_dict[lib_consts.ID] = self.listener2_id
        self.test_listener2_dict[lib_consts.NAME] = 'listener_2'
        self.test_listener2_dict[lib_consts.DESCRIPTION] = 'Listener 1'
        self.test_listener2_dict[lib_consts.DEFAULT_POOL_ID] = self.pool2_id
        self.test_listener2_dict[
            lib_consts.DEFAULT_POOL] = self.test_pool2_dict
        del self.test_listener2_dict[lib_consts.L7POLICIES]
        del self.test_listener2_dict[constants.SNI_CONTAINERS]
        del self.test_listener2_dict[constants.CLIENT_CA_TLS_CERTIFICATE_ID]
        del self.test_listener2_dict[constants.CLIENT_CRL_CONTAINER_ID]

        self.test_listeners = [self.test_listener1_dict,
                               self.test_listener2_dict]

        self.db_listener1 = data_models.Listener(**self.test_listener1_dict)
        self.db_listener2 = data_models.Listener(**self.test_listener2_dict)
        self.db_listener1.default_pool = self.db_pool1
        self.db_listener2.default_pool = self.db_pool2
        self.db_listener1.l7policies = self.db_l7policies
        self.db_listener1.sni_containers = [
            data_models.SNI(tls_container_id='2'),
            data_models.SNI(tls_container_id='3')]

        self.test_db_listeners = [self.db_listener1, self.db_listener2]

        cert1 = data_models.TLSContainer(certificate='cert 1')
        cert2 = data_models.TLSContainer(certificate='cert 2')
        cert3 = data_models.TLSContainer(certificate='cert 3')
        ca_cert = 'ca cert'
        crl_file_content = 'X509 CRL FILE'

        self.provider_listener1_dict = {
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.ALLOWED_CIDRS: ['192.0.2.0/24', '198.51.100.0/24'],
            lib_consts.ALPN_PROTOCOLS: [lib_consts.ALPN_PROTOCOL_HTTP_1_1],
            lib_consts.CONNECTION_LIMIT: 10000,
            lib_consts.DEFAULT_POOL: self.provider_pool1_dict,
            lib_consts.DEFAULT_POOL_ID: self.pool1_id,
            lib_consts.DEFAULT_TLS_CONTAINER_DATA: cert1.to_dict(),
            lib_consts.DEFAULT_TLS_CONTAINER_REF:
                self.default_tls_container_ref,
            lib_consts.DESCRIPTION: 'Listener 1',
            lib_consts.INSERT_HEADERS: {},
            lib_consts.L7POLICIES: self.provider_l7policies_dict,
            lib_consts.LISTENER_ID: self.listener1_id,
            lib_consts.LOADBALANCER_ID: self.lb_id,
            lib_consts.NAME: 'listener_1',
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.PROTOCOL: lib_consts.PROTOCOL_TCP,
            lib_consts.PROTOCOL_PORT: 90,
            lib_consts.SNI_CONTAINER_DATA: [cert2.to_dict(), cert3.to_dict()],
            lib_consts.SNI_CONTAINER_REFS: [self.sni_container_ref_1,
                                            self.sni_container_ref_2],
            lib_consts.TIMEOUT_CLIENT_DATA: 1000,
            lib_consts.TIMEOUT_MEMBER_CONNECT: 2000,
            lib_consts.TIMEOUT_MEMBER_DATA: 3000,
            lib_consts.TIMEOUT_TCP_INSPECT: 4000,
            lib_consts.CLIENT_CA_TLS_CONTAINER_REF:
                self.client_ca_tls_certificate_ref,
            lib_consts.CLIENT_CA_TLS_CONTAINER_DATA: ca_cert,
            lib_consts.CLIENT_AUTHENTICATION: constants.CLIENT_AUTH_NONE,
            lib_consts.CLIENT_CRL_CONTAINER_REF: self.client_crl_container_ref,
            lib_consts.CLIENT_CRL_CONTAINER_DATA: crl_file_content,
            lib_consts.TLS_CIPHERS: constants.CIPHERS_OWASP_SUITE_B,
            lib_consts.TLS_VERSIONS: constants.TLS_VERSIONS_OWASP_SUITE_B,
            lib_consts.ALPN_PROTOCOLS:
                constants.AMPHORA_SUPPORTED_ALPN_PROTOCOLS
        }

        self.provider_listener2_dict = copy.deepcopy(
            self.provider_listener1_dict)
        self.provider_listener2_dict[
            lib_consts.LISTENER_ID] = self.listener2_id
        self.provider_listener2_dict[lib_consts.NAME] = 'listener_2'
        self.provider_listener2_dict[lib_consts.DESCRIPTION] = 'Listener 1'
        self.provider_listener2_dict[
            lib_consts.DEFAULT_POOL_ID] = self.pool2_id
        self.provider_listener2_dict[
            lib_consts.DEFAULT_POOL] = self.provider_pool2_dict
        del self.provider_listener2_dict[lib_consts.L7POLICIES]
        self.provider_listener2_dict[
            lib_consts.CLIENT_CA_TLS_CONTAINER_REF] = None
        del self.provider_listener2_dict[
            lib_consts.CLIENT_CA_TLS_CONTAINER_DATA]
        self.provider_listener2_dict[lib_consts.CLIENT_AUTHENTICATION] = (
            constants.CLIENT_AUTH_NONE)
        self.provider_listener2_dict[
            lib_consts.CLIENT_CRL_CONTAINER_REF] = None
        del self.provider_listener2_dict[lib_consts.CLIENT_CRL_CONTAINER_DATA]

        self.provider_listener1 = driver_dm.Listener(
            **self.provider_listener1_dict)
        self.provider_listener2 = driver_dm.Listener(
            **self.provider_listener2_dict)
        self.provider_listener1.default_pool = self.provider_pool1
        self.provider_listener2.default_pool = self.provider_pool2
        self.provider_listener1.l7policies = self.provider_l7policies

        self.provider_listeners = [self.provider_listener1,
                                   self.provider_listener2]

        self.test_vip_dict = {constants.IP_ADDRESS: self.ip_address,
                              constants.NETWORK_ID: self.network_id,
                              constants.PORT_ID: self.port_id,
                              lib_consts.SUBNET_ID: self.subnet_id,
                              constants.QOS_POLICY_ID: self.qos_policy_id,
                              constants.OCTAVIA_OWNED: None}

        self.provider_vip_dict = {
            lib_consts.VIP_ADDRESS: self.ip_address,
            lib_consts.VIP_NETWORK_ID: self.network_id,
            lib_consts.VIP_PORT_ID: self.port_id,
            lib_consts.VIP_SUBNET_ID: self.subnet_id,
            lib_consts.VIP_QOS_POLICY_ID: self.qos_policy_id,
            constants.OCTAVIA_OWNED: None}

        self.db_vip = data_models.Vip(
            ip_address=self.ip_address,
            network_id=self.network_id,
            port_id=self.port_id,
            subnet_id=self.subnet_id,
            qos_policy_id=self.qos_policy_id)

        self.test_loadbalancer1_dict = {
            lib_consts.NAME: self.lb_name,
            lib_consts.DESCRIPTION: self.lb_description,
            constants.ENABLED: True,
            lib_consts.PROVISIONING_STATUS: lib_consts.PENDING_UPDATE,
            lib_consts.OPERATING_STATUS: lib_consts.OFFLINE,
            constants.TOPOLOGY: constants.TOPOLOGY_ACTIVE_STANDBY,
            constants.VRRP_GROUP: None,
            constants.PROVIDER: constants.AMPHORA,
            constants.SERVER_GROUP_ID: uuidutils.generate_uuid(),
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.ID: self.lb_id, constants.FLAVOR_ID: self.flavor_id,
            constants.TAGS: ['test_tag']}

        self.provider_loadbalancer_dict = {
            lib_consts.ADDITIONAL_VIPS: None,
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.DESCRIPTION: self.lb_description,
            lib_consts.FLAVOR: {"something": "else"},
            lib_consts.LISTENERS: None,
            lib_consts.LOADBALANCER_ID: self.lb_id,
            lib_consts.NAME: self.lb_name,
            lib_consts.POOLS: None,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.VIP_ADDRESS: self.ip_address,
            lib_consts.VIP_NETWORK_ID: self.network_id,
            lib_consts.VIP_PORT_ID: self.port_id,
            lib_consts.VIP_QOS_POLICY_ID: self.qos_policy_id,
            lib_consts.VIP_SUBNET_ID: self.subnet_id}

        self.provider_loadbalancer_tree_dict = {
            lib_consts.ADDITIONAL_VIPS: None,
            lib_consts.ADMIN_STATE_UP: True,
            lib_consts.AVAILABILITY_ZONE: None,
            lib_consts.DESCRIPTION: self.lb_description,
            lib_consts.FLAVOR: {"something": "else"},
            lib_consts.LISTENERS: None,
            lib_consts.LOADBALANCER_ID: self.lb_id,
            lib_consts.NAME: self.lb_name,
            lib_consts.POOLS: None,
            lib_consts.PROJECT_ID: self.project_id,
            lib_consts.VIP_ADDRESS: self.ip_address,
            lib_consts.VIP_NETWORK_ID: self.network_id,
            lib_consts.VIP_PORT_ID: self.port_id,
            lib_consts.VIP_QOS_POLICY_ID: self.qos_policy_id,
            lib_consts.VIP_SUBNET_ID: self.subnet_id}



#    Copyright 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc

from oslo_config import cfg

CONF = cfg.CONF


class LvsListenerApiServerBase(object, metaclass=abc.ABCMeta):
    """Base LVS Listener Server API

    """

    _SUBSCRIBED_AMP_COMPILE = []

    def get_subscribed_amp_compile_info(self):
        return self._SUBSCRIBED_AMP_COMPILE

    @abc.abstractmethod
    def upload_lvs_listener_config(self, listener_id):
        """Upload the configuration for LVS.

        :param listener_id: The id of a LVS Listener

        :returns: HTTP response with status code.
        :raises Exception: If any file / directory is not found or
                            fail to create.

        """

    @abc.abstractmethod
    def get_lvs_listener_config(self, listener_id):
        """Gets the LVS Listener configuration details

        :param listener_id: the id of the LVS Listener

        :returns: HTTP response with status code.
        :raises Exception: If the listener is failed to find.

        """

    @abc.abstractmethod
    def manage_lvs_listener(self, listener_id, action):
        """Gets the LVS Listener configuration details

        :param listener_id: the id of the LVS Listener
        :param action: the operation type.

        :returns: HTTP response with status code.
        :raises Exception: If the listener is failed to find.

        """

    @abc.abstractmethod
    def get_all_lvs_listeners_status(self):
        """Gets the status of all LVS Listeners

        This method will not consult the stats socket
        so a listener might show as ACTIVE but still be
        in ERROR

        :returns: a list of LVS Listener status
        :raises Exception: If the listener pid located directory is not exist

        """

    @abc.abstractmethod
    def delete_lvs_listener(self, listener_id):
        """Delete a LVS Listener from a amphora

        :param listener_id: The id of the listener

        :returns: HTTP response with status code.
        :raises Exception: If unsupport initial system of amphora.

        """



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from taskflow.patterns import linear_flow

from octavia.common import constants
from octavia.controller.worker.v2.tasks import amphora_driver_tasks
from octavia.controller.worker.v2.tasks import database_tasks
from octavia.controller.worker.v2.tasks import lifecycle_tasks


class HealthMonitorFlows(object):

    def get_create_health_monitor_flow(self):
        """Create a flow to create a health monitor

        :returns: The flow for creating a health monitor
        """
        create_hm_flow = linear_flow.Flow(constants.CREATE_HEALTH_MONITOR_FLOW)
        create_hm_flow.add(lifecycle_tasks.HealthMonitorToErrorOnRevertTask(
            requires=[constants.HEALTH_MON,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        create_hm_flow.add(database_tasks.MarkHealthMonitorPendingCreateInDB(
            requires=constants.HEALTH_MON))
        create_hm_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        create_hm_flow.add(database_tasks.MarkHealthMonitorActiveInDB(
            requires=constants.HEALTH_MON))
        create_hm_flow.add(database_tasks.MarkPoolActiveInDB(
            requires=constants.POOL_ID))
        create_hm_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=(constants.LOADBALANCER_ID, constants.LISTENERS)))

        return create_hm_flow

    def get_delete_health_monitor_flow(self):
        """Create a flow to delete a health monitor

        :returns: The flow for deleting a health monitor
        """
        delete_hm_flow = linear_flow.Flow(constants.DELETE_HEALTH_MONITOR_FLOW)
        delete_hm_flow.add(lifecycle_tasks.HealthMonitorToErrorOnRevertTask(
            requires=[constants.HEALTH_MON,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        delete_hm_flow.add(database_tasks.MarkHealthMonitorPendingDeleteInDB(
            requires=constants.HEALTH_MON))
        delete_hm_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        delete_hm_flow.add(database_tasks.DeleteHealthMonitorInDB(
            requires=constants.HEALTH_MON))
        delete_hm_flow.add(database_tasks.DecrementHealthMonitorQuota(
            requires=constants.PROJECT_ID))
        delete_hm_flow.add(
            database_tasks.UpdatePoolMembersOperatingStatusInDB(
                requires=constants.POOL_ID,
                inject={constants.OPERATING_STATUS: constants.NO_MONITOR}))
        delete_hm_flow.add(database_tasks.MarkPoolActiveInDB(
            requires=constants.POOL_ID))
        delete_hm_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=(constants.LOADBALANCER_ID, constants.LISTENERS)))

        return delete_hm_flow

    def get_update_health_monitor_flow(self):
        """Create a flow to update a health monitor

        :returns: The flow for updating a health monitor
        """
        update_hm_flow = linear_flow.Flow(constants.UPDATE_HEALTH_MONITOR_FLOW)
        update_hm_flow.add(lifecycle_tasks.HealthMonitorToErrorOnRevertTask(
            requires=[constants.HEALTH_MON,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        update_hm_flow.add(database_tasks.MarkHealthMonitorPendingUpdateInDB(
            requires=constants.HEALTH_MON))
        update_hm_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        update_hm_flow.add(database_tasks.UpdateHealthMonInDB(
            requires=[constants.HEALTH_MON, constants.UPDATE_DICT]))
        update_hm_flow.add(database_tasks.MarkHealthMonitorActiveInDB(
            requires=constants.HEALTH_MON))
        update_hm_flow.add(database_tasks.MarkPoolActiveInDB(
            requires=constants.POOL_ID))
        update_hm_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=(constants.LOADBALANCER_ID, constants.LISTENERS)))

        return update_hm_flow



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime

from oslo_config import cfg
from oslo_middleware.healthcheck import pluginbase

from octavia.db import api as db_apis
from octavia.db import healthcheck

CONF = cfg.CONF


class OctaviaDBHealthcheck(pluginbase.HealthcheckBaseExtension):

    UNAVAILABLE_REASON = 'The Octavia database is unavailable.'

    last_check = None
    last_result = None
    last_message = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def healthcheck(self, server_port):
        try:
            if (self.last_check is not None and
                    ((datetime.datetime.now() -
                      self.last_check).total_seconds()) <
                    CONF.api_settings.healthcheck_refresh_interval):
                result = self.last_result
                message = self.last_message
            else:
                result, message = healthcheck.check_database_connection(
                    db_apis.get_session())
                self.last_check = datetime.datetime.now()
                self.last_result = result
                self.last_message = message
            if result:
                return OctaviaDBCheckResult(available=True, reason="OK")
            else:
                return OctaviaDBCheckResult(available=False,
                                            reason=self.UNAVAILABLE_REASON,
                                            details=message)
        except Exception as e:
            return OctaviaDBCheckResult(available=False,
                                        reason=self.UNAVAILABLE_REASON,
                                        details=str(e))


class OctaviaDBCheckResult(pluginbase.HealthcheckResult):
    """Result sub-class to provide a unique name in detail reports."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ipaddress
from unittest import mock

from octavia.amphorae.backends.utils import interface_file
from octavia.common import constants as consts
from octavia.tests.common import utils as test_utils
import octavia.tests.unit.base as base


class TestInterfaceFile(base.TestCase):
    def test_vip_interface_file(self):
        netns_interface = 'eth1234'
        MTU = 1450
        VIP_ADDRESS = '192.0.2.2'
        SUBNET_CIDR = '192.0.2.0/24'
        GATEWAY = '192.0.2.1'
        DEST1 = '198.51.100.0/24'
        NEXTHOP = '192.0.2.1'
        VRRP_IP_ADDRESS = '192.10.2.4'
        TOPOLOGY = 'SINGLE'

        cidr = ipaddress.ip_network(SUBNET_CIDR)
        prefixlen = cidr.prefixlen

        vip_interface_file = interface_file.VIPInterfaceFile(
            name=netns_interface,
            mtu=MTU,
            vip=VIP_ADDRESS,
            ip_version=cidr.version,
            prefixlen=prefixlen,
            gateway=GATEWAY,
            vrrp_ip=VRRP_IP_ADDRESS,
            host_routes=[
                {'destination': DEST1, 'nexthop': NEXTHOP}
            ],
            topology=TOPOLOGY)

        expected_dict = {
            consts.NAME: netns_interface,
            consts.MTU: MTU,
            consts.ADDRESSES: [
                {
                    consts.ADDRESS: VRRP_IP_ADDRESS,
                    consts.PREFIXLEN: prefixlen
                },
                {
                    consts.ADDRESS: VIP_ADDRESS,
                    consts.PREFIXLEN: prefixlen
                }
            ],
            consts.ROUTES: [
                {
                    consts.DST: "0.0.0.0/0",
                    consts.GATEWAY: GATEWAY,
                    consts.FLAGS: [consts.ONLINK],
                },
                {
                    consts.DST: "0.0.0.0/0",
                    consts.GATEWAY: GATEWAY,
                    consts.FLAGS: [consts.ONLINK],
                    consts.TABLE: 1
                },
                {
                    consts.DST: cidr.exploded,
                    consts.PREFSRC: VIP_ADDRESS,
                    consts.SCOPE: 'link',
                    consts.TABLE: 1
                },
                {
                    consts.DST: DEST1,
                    consts.GATEWAY: NEXTHOP,
                    consts.FLAGS: [consts.ONLINK]
                },
                {
                    consts.DST: DEST1,
                    consts.GATEWAY: NEXTHOP,
                    consts.TABLE: 1,
                    consts.FLAGS: [consts.ONLINK]
                }
            ],
            consts.RULES: [
                {
                    consts.SRC: VIP_ADDRESS,
                    consts.SRC_LEN: 32,
                    consts.TABLE: 1
                }
            ],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv4 "
                        "{}".format(netns_interface))
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv4 "
                        "{}".format(netns_interface))
                }]
            }
        }

        with mock.patch('os.open'), mock.patch('os.fdopen'), mock.patch(
                'octavia.amphorae.backends.utils.interface_file.'
                'InterfaceFile.dump') as mock_dump:
            vip_interface_file.write()

            mock_dump.assert_called_once()

            args = mock_dump.mock_calls[0][1]
            test_utils.assert_interface_files_equal(
                self, expected_dict, args[0])

    def test_vip_interface_file_dhcp(self):
        netns_interface = 'eth1234'
        MTU = 1450
        VIP_ADDRESS = '192.0.2.2'
        SUBNET_CIDR = '192.0.2.0/24'
        TOPOLOGY = 'SINGLE'

        cidr = ipaddress.ip_network(SUBNET_CIDR)
        prefixlen = cidr.prefixlen

        vip_interface_file = interface_file.VIPInterfaceFile(
            name=netns_interface,
            mtu=MTU,
            vip=VIP_ADDRESS,
            ip_version=cidr.version,
            prefixlen=prefixlen,
            gateway=None,
            vrrp_ip=None,
            host_routes=[],
            topology=TOPOLOGY)

        expected_dict = {
            consts.NAME: netns_interface,
            consts.MTU: MTU,
            consts.ADDRESSES: [
                {
                    consts.DHCP: True
                }, {
                    consts.ADDRESS: VIP_ADDRESS,
                    consts.PREFIXLEN: prefixlen
                }
            ],
            consts.ROUTES: [
                {
                    consts.DST: cidr.exploded,
                    consts.PREFSRC: VIP_ADDRESS,
                    consts.SCOPE: 'link',
                    consts.TABLE: 1
                }
            ],
            consts.RULES: [
                {
                    consts.SRC: VIP_ADDRESS,
                    consts.SRC_LEN: 32,
                    consts.TABLE: 1
                }
            ],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv4 "
                        "{}".format(netns_interface))
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv4 "
                        "{}".format(netns_interface))
                }]
            }
        }

        with mock.patch('os.open'), mock.patch('os.fdopen'), mock.patch(
                'octavia.amphorae.backends.utils.interface_file.'
                'InterfaceFile.dump') as mock_dump:
            vip_interface_file.write()

            mock_dump.assert_called_once()

            args = mock_dump.mock_calls[0][1]
            test_utils.assert_interface_files_equal(
                self, expected_dict, args[0])

    def test_vip_interface_file_active_standby(self):
        netns_interface = 'eth1234'
        MTU = 1450
        VIP_ADDRESS = '192.0.2.2'
        SUBNET_CIDR = '192.0.2.0/24'
        GATEWAY = '192.0.2.1'
        VRRP_IP_ADDRESS = '192.10.2.4'
        TOPOLOGY = 'ACTIVE_STANDBY'

        cidr = ipaddress.ip_network(SUBNET_CIDR)
        prefixlen = cidr.prefixlen

        vip_interface_file = interface_file.VIPInterfaceFile(
            name=netns_interface,
            mtu=MTU,
            vip=VIP_ADDRESS,
            ip_version=cidr.version,
            prefixlen=prefixlen,
            gateway=GATEWAY,
            vrrp_ip=VRRP_IP_ADDRESS,
            host_routes=[],
            topology=TOPOLOGY)

        expected_dict = {
            consts.NAME: netns_interface,
            consts.MTU: MTU,
            consts.ADDRESSES: [
                {
                    consts.ADDRESS: VRRP_IP_ADDRESS,
                    consts.PREFIXLEN: prefixlen
                }
            ],
            consts.ROUTES: [
                {
                    consts.DST: "0.0.0.0/0",
                    consts.GATEWAY: GATEWAY,
                    consts.FLAGS: [consts.ONLINK],
                },
                {
                    consts.DST: "0.0.0.0/0",
                    consts.GATEWAY: GATEWAY,
                    consts.FLAGS: [consts.ONLINK],
                    consts.TABLE: 1
                }
            ],
            consts.RULES: [],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv4 "
                        "{}".format(netns_interface))
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv4 "
                        "{}".format(netns_interface))
                }]
            }
        }

        with mock.patch('os.open'), mock.patch('os.fdopen'), mock.patch(
                'octavia.amphorae.backends.utils.interface_file.'
                'InterfaceFile.dump') as mock_dump:
            vip_interface_file.write()

            mock_dump.assert_called_once()

            args = mock_dump.mock_calls[0][1]
            test_utils.assert_interface_files_equal(
                self, expected_dict, args[0])

    def test_vip_interface_file_ipv6(self):
        netns_interface = 'eth1234'
        MTU = 1450
        VIP_ADDRESS = '2001:db8::7'
        SUBNET_CIDR = '2001:db8::/64'
        GATEWAY = '2001:db8::1'
        DEST1 = '2001:db8:2::/64'
        NEXTHOP = '2001:db8:2::1'
        VRRP_IP_ADDRESS = '2001:db8::42'
        TOPOLOGY = 'SINGLE'

        cidr = ipaddress.ip_network(SUBNET_CIDR)
        prefixlen = cidr.prefixlen

        vip_interface_file = interface_file.VIPInterfaceFile(
            name=netns_interface,
            mtu=MTU,
            vip=VIP_ADDRESS,
            ip_version=cidr.version,
            prefixlen=prefixlen,
            gateway=GATEWAY,
            vrrp_ip=VRRP_IP_ADDRESS,
            host_routes=[
                {'destination': DEST1, 'nexthop': NEXTHOP}
            ],
            topology=TOPOLOGY)

        expected_dict = {
            consts.NAME: netns_interface,
            consts.MTU: MTU,
            consts.ADDRESSES: [
                {
                    consts.ADDRESS: VRRP_IP_ADDRESS,
                    consts.PREFIXLEN: prefixlen
                },
                {
                    consts.ADDRESS: VIP_ADDRESS,
                    consts.PREFIXLEN: prefixlen
                }
            ],
            consts.ROUTES: [
                {
                    consts.DST: "::/0",
                    consts.GATEWAY: GATEWAY,
                    consts.FLAGS: [consts.ONLINK]
                },
                {
                    consts.DST: "::/0",
                    consts.GATEWAY: GATEWAY,
                    consts.TABLE: 1,
                    consts.FLAGS: [consts.ONLINK]
                },
                {
                    consts.DST: cidr.exploded,
                    consts.PREFSRC: VIP_ADDRESS,
                    consts.SCOPE: 'link',
                    consts.TABLE: 1
                },
                {
                    consts.DST: DEST1,
                    consts.GATEWAY: NEXTHOP,
                    consts.FLAGS: [consts.ONLINK]
                },
                {
                    consts.DST: DEST1,
                    consts.GATEWAY: NEXTHOP,
                    consts.TABLE: 1,
                    consts.FLAGS: [consts.ONLINK]
                }
            ],
            consts.RULES: [
                {
                    consts.SRC: VIP_ADDRESS,
                    consts.SRC_LEN: 128,
                    consts.TABLE: 1
                }
            ],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv6 "
                        "{}".format(netns_interface))
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv6 "
                        "{}".format(netns_interface))
                }]
            }
        }

        with mock.patch('os.open'), mock.patch('os.fdopen'), mock.patch(
                'octavia.amphorae.backends.utils.interface_file.'
                'InterfaceFile.dump') as mock_dump:
            vip_interface_file.write()

            mock_dump.assert_called_once()

            args = mock_dump.mock_calls[0][1]
            test_utils.assert_interface_files_equal(
                self, expected_dict, args[0])

    def test_port_interface_file(self):
        netns_interface = 'eth1234'
        FIXED_IP = '192.0.2.2'
        SUBNET_CIDR = '192.0.2.0/24'
        DEST1 = '198.51.100.0/24'
        DEST2 = '203.0.113.0/24'
        DEST3 = 'fd01::/64'
        NEXTHOP = '192.0.2.1'
        NEXTHOP2 = '2001:db7::8'
        MTU = 1450
        FIXED_IP_IPV6 = '2001:0db8:0000:0000:0000:0000:0000:0001'
        SUBNET_CIDR_IPV6 = '2001:db8::/64'
        fixed_ips = [{'ip_address': FIXED_IP,
                      'subnet_cidr': SUBNET_CIDR,
                      'host_routes': [
                          {'destination': DEST1, 'nexthop': NEXTHOP},
                          {'destination': DEST2, 'nexthop': NEXTHOP}
                      ]},
                     {'ip_address': FIXED_IP_IPV6,
                      'subnet_cidr': SUBNET_CIDR_IPV6,
                      'host_routes': [
                          {'destination': DEST3, 'nexthop': NEXTHOP2}
                      ]},
                     ]

        port_interface_file = interface_file.PortInterfaceFile(
            name=netns_interface,
            fixed_ips=fixed_ips,
            mtu=MTU)

        expected_dict = {
            consts.NAME: netns_interface,
            consts.MTU: MTU,
            consts.ADDRESSES: [
                {
                    consts.ADDRESS: FIXED_IP,
                    consts.PREFIXLEN: (
                        ipaddress.ip_network(SUBNET_CIDR).prefixlen
                    )
                },
                {
                    consts.ADDRESS: FIXED_IP_IPV6,
                    consts.PREFIXLEN: (
                        ipaddress.ip_network(SUBNET_CIDR_IPV6).prefixlen
                    )
                }
            ],
            consts.ROUTES: [
                {
                    consts.DST: DEST1,
                    consts.GATEWAY: NEXTHOP
                },
                {
                    consts.DST: DEST2,
                    consts.GATEWAY: NEXTHOP
                },
                {
                    consts.DST: DEST3,
                    consts.GATEWAY: NEXTHOP2
                }
            ],
            consts.RULES: [],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv4 "
                        "{}".format(netns_interface))
                }, {
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv6 "
                        "{}".format(netns_interface))
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv4 "
                        "{}".format(netns_interface))
                }, {
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv6 "
                        "{}".format(netns_interface))
                }]
            }
        }

        with mock.patch('os.open'), mock.patch('os.fdopen'), mock.patch(
                'octavia.amphorae.backends.utils.interface_file.'
                'InterfaceFile.dump') as mock_dump:
            port_interface_file.write()

            mock_dump.assert_called_once()

            args = mock_dump.mock_calls[0][1]
            test_utils.assert_interface_files_equal(
                self, expected_dict, args[0])

    def test_port_interface_file_dhcp(self):
        netns_interface = 'eth1234'
        MTU = 1450

        port_interface_file = interface_file.PortInterfaceFile(
            name=netns_interface,
            fixed_ips=None,
            mtu=MTU)

        expected_dict = {
            consts.NAME: netns_interface,
            consts.MTU: MTU,
            consts.ADDRESSES: [{
                consts.DHCP: True,
                consts.IPV6AUTO: True,
            }],
            consts.ROUTES: [],
            consts.RULES: [],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv4 "
                        "{}".format(netns_interface))
                }, {
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh add ipv6 "
                        "{}".format(netns_interface))
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv4 "
                        "{}".format(netns_interface))
                }, {
                    consts.COMMAND: (
                        "/usr/local/bin/lvs-masquerade.sh delete ipv6 "
                        "{}".format(netns_interface))
                }]
            }
        }

        with mock.patch('os.open'), mock.patch('os.fdopen'), mock.patch(
                'octavia.amphorae.backends.utils.interface_file.'
                'InterfaceFile.dump') as mock_dump:
            port_interface_file.write()

            mock_dump.assert_called_once()

            args = mock_dump.mock_calls[0][1]
            test_utils.assert_interface_files_equal(
                self, expected_dict, args[0])

    def test_from_file(self):
        filename = 'interface.json'
        content = ('{"addresses": [\n'
                   '{"address": "10.0.0.181",\n'
                   '"prefixlen": 26}\n'
                   '],\n'
                   '"mtu": 1450,\n'
                   '"name": "eth1",\n'
                   '"routes": [\n'
                   '{"dst": "0.0.0.0/0",\n'
                   '"gateway": "10.0.0.129",\n'
                   '"onlink": true}\n'
                   '],\n'
                   '"rules": [\n'
                   '{"src": "10.0.0.157",\n'
                   '"src_len": 32,\n'
                   '"table": 1}\n'
                   '],\n'
                   '"scripts": {\n'
                   '"down": [\n'
                   '{"command": "script-down"}\n'
                   '], "up": [ \n'
                   '{"command": "script-up"}\n'
                   ']}}\n')

        self.useFixture(
            test_utils.OpenFixture(filename,
                                   contents=content))

        iface = interface_file.InterfaceFile.from_file(filename)

        expected_dict = {
            consts.NAME: "eth1",
            consts.MTU: 1450,
            consts.ADDRESSES: [{
                consts.ADDRESS: "10.0.0.181",
                consts.PREFIXLEN: 26
            }],
            consts.ROUTES: [{
                consts.DST: "0.0.0.0/0",
                consts.GATEWAY: "10.0.0.129",
                consts.FLAGS: [consts.ONLINK]
            }],
            consts.RULES: [{
                consts.SRC: "10.0.0.157",
                consts.SRC_LEN: 32,
                consts.TABLE: 1
            }],
            consts.SCRIPTS: {
                consts.IFACE_UP: [{
                    consts.COMMAND: "script-up"
                }],
                consts.IFACE_DOWN: [{
                    consts.COMMAND: "script-down"
                }]
            }
        }

        self.assertEqual(expected_dict[consts.NAME], iface.name)
        self.assertEqual(expected_dict[consts.MTU], iface.mtu)
        test_utils.assert_address_lists_equal(
            self, expected_dict[consts.ADDRESSES], iface.addresses)
        test_utils.assert_rule_lists_equal(
            self, expected_dict[consts.RULES], iface.rules)
        test_utils.assert_script_lists_equal(
            self, expected_dict[consts.SCRIPTS], iface.scripts)



#    Copyright 2014 Rackspace
#    Copyright 2016 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia_lib.api.drivers import data_models as driver_dm
from octavia_lib.common import constants as lib_consts
from oslo_config import cfg
from oslo_db import exception as odb_exceptions
from oslo_log import log as logging
from oslo_utils import excutils
from pecan import request as pecan_request
from wsme import types as wtypes
from wsmeext import pecan as wsme_pecan

from octavia.api.drivers import driver_factory
from octavia.api.drivers import utils as driver_utils
from octavia.api.v2.controllers import base
from octavia.api.v2.types import health_monitor as hm_types
from octavia.common import constants as consts
from octavia.common import data_models
from octavia.common import exceptions
from octavia.db import api as db_api
from octavia.db import prepare as db_prepare
from octavia.i18n import _


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class HealthMonitorController(base.BaseController):
    RBAC_TYPE = consts.RBAC_HEALTHMONITOR

    def __init__(self):
        super().__init__()

    @wsme_pecan.wsexpose(hm_types.HealthMonitorRootResponse, wtypes.text,
                         [wtypes.text], ignore_extra_args=True)
    def get_one(self, id, fields=None):
        """Gets a single healthmonitor's details."""
        context = pecan_request.context.get('octavia_context')
        db_hm = self._get_db_hm(context.session, id, show_deleted=False)

        self._auth_validate_action(context, db_hm.project_id,
                                   consts.RBAC_GET_ONE)

        result = self._convert_db_to_type(
            db_hm, hm_types.HealthMonitorResponse)
        if fields is not None:
            result = self._filter_fields([result], fields)[0]
        return hm_types.HealthMonitorRootResponse(healthmonitor=result)

    @wsme_pecan.wsexpose(hm_types.HealthMonitorsRootResponse, wtypes.text,
                         [wtypes.text], ignore_extra_args=True)
    def get_all(self, project_id=None, fields=None):
        """Gets all health monitors."""
        pcontext = pecan_request.context
        context = pcontext.get('octavia_context')

        query_filter = self._auth_get_all(context, project_id)

        db_hm, links = self.repositories.health_monitor.get_all_API_list(
            context.session, show_deleted=False,
            pagination_helper=pcontext.get(consts.PAGINATION_HELPER),
            **query_filter)
        result = self._convert_db_to_type(
            db_hm, [hm_types.HealthMonitorResponse])
        if fields is not None:
            result = self._filter_fields(result, fields)
        return hm_types.HealthMonitorsRootResponse(
            healthmonitors=result, healthmonitors_links=links)

    def _get_affected_listener_ids(self, session, hm):
        """Gets a list of all listeners this request potentially affects."""
        pool = self.repositories.pool.get(session, id=hm.pool_id)
        listener_ids = [li.id for li in pool.listeners]
        return listener_ids

    def _test_lb_and_listener_and_pool_statuses(self, session, hm):
        """Verify load balancer is in a mutable state."""
        # We need to verify that any listeners referencing this pool are also
        # mutable
        pool = self.repositories.pool.get(session, id=hm.pool_id)
        load_balancer_id = pool.load_balancer_id
        # Check the parent is not locked for some reason (ERROR, etc.)
        if pool.provisioning_status not in consts.MUTABLE_STATUSES:
            raise exceptions.ImmutableObject(resource='Pool', id=hm.pool_id)
        if not self.repositories.test_and_set_lb_and_listeners_prov_status(
                session, load_balancer_id,
                consts.PENDING_UPDATE, consts.PENDING_UPDATE,
                listener_ids=self._get_affected_listener_ids(session, hm),
                pool_id=hm.pool_id):
            LOG.info("Health Monitor cannot be created or modified because "
                     "the Load Balancer is in an immutable state")
            raise exceptions.ImmutableObject(resource='Load Balancer',
                                             id=load_balancer_id)

    def _validate_create_hm(self, lock_session, hm_dict):
        """Validate creating health monitor on pool."""
        mandatory_fields = (consts.TYPE, consts.DELAY, consts.TIMEOUT,
                            consts.POOL_ID)
        for field in mandatory_fields:
            if hm_dict.get(field, None) is None:
                raise exceptions.InvalidOption(value='None', option=field)
        # MAX_RETRIES is renamed fall_threshold so handle is special
        if hm_dict.get(consts.RISE_THRESHOLD, None) is None:
            raise exceptions.InvalidOption(value='None',
                                           option=consts.MAX_RETRIES)

        if hm_dict[consts.TYPE] not in (consts.HEALTH_MONITOR_HTTP,
                                        consts.HEALTH_MONITOR_HTTPS):
            if hm_dict.get(consts.HTTP_METHOD, None):
                raise exceptions.InvalidOption(
                    value=consts.HTTP_METHOD, option='health monitors of '
                    'type {}'.format(hm_dict[consts.TYPE]))
            if hm_dict.get(consts.URL_PATH, None):
                raise exceptions.InvalidOption(
                    value=consts.URL_PATH, option='health monitors of '
                    'type {}'.format(hm_dict[consts.TYPE]))
            if hm_dict.get(consts.EXPECTED_CODES, None):
                raise exceptions.InvalidOption(
                    value=consts.EXPECTED_CODES, option='health monitors of '
                    'type {}'.format(hm_dict[consts.TYPE]))
        else:
            if not hm_dict.get(consts.HTTP_METHOD, None):
                hm_dict[consts.HTTP_METHOD] = (
                    consts.HEALTH_MONITOR_HTTP_DEFAULT_METHOD)
            if not hm_dict.get(consts.URL_PATH, None):
                hm_dict[consts.URL_PATH] = (
                    consts.HEALTH_MONITOR_DEFAULT_URL_PATH)
            if not hm_dict.get(consts.EXPECTED_CODES, None):
                hm_dict[consts.EXPECTED_CODES] = (
                    consts.HEALTH_MONITOR_DEFAULT_EXPECTED_CODES)

        if hm_dict.get('domain_name') and not hm_dict.get('http_version'):
            raise exceptions.ValidationException(
                detail=_("'http_version' must be specified when 'domain_name' "
                         "is provided."))

        if hm_dict.get('http_version') and hm_dict.get('domain_name'):
            if hm_dict['http_version'] < 1.1:
                raise exceptions.InvalidOption(
                    value='http_version %s' % hm_dict['http_version'],
                    option='health monitors HTTP 1.1 domain name health check')

        try:
            return self.repositories.health_monitor.create(
                lock_session, **hm_dict)
        except odb_exceptions.DBDuplicateEntry as e:
            raise exceptions.DuplicateHealthMonitor() from e
        except odb_exceptions.DBError as e:
            # TODO(blogan): will have to do separate validation protocol
            # before creation or update since the exception messages
            # do not give any information as to what constraint failed
            raise exceptions.InvalidOption(value='', option='') from e

    def _validate_healthmonitor_request_for_udp_sctp(self, request,
                                                     pool_protocol):
        if request.type not in (
                consts.HEALTH_MONITOR_UDP_CONNECT,
                lib_consts.HEALTH_MONITOR_SCTP,
                consts.HEALTH_MONITOR_TCP,
                consts.HEALTH_MONITOR_HTTP):
            raise exceptions.ValidationException(detail=_(
                "The associated pool protocol is %(pool_protocol)s, so only "
                "a %(types)s health monitor is supported.") % {
                    'pool_protocol': pool_protocol,
                    'types': '/'.join((consts.HEALTH_MONITOR_UDP_CONNECT,
                                       lib_consts.HEALTH_MONITOR_SCTP,
                                       consts.HEALTH_MONITOR_TCP,
                                       consts.HEALTH_MONITOR_HTTP))})
        # check the delay value if the HM type is UDP-CONNECT
        hm_is_type_udp = (
            request.type == consts.HEALTH_MONITOR_UDP_CONNECT)
        conf_min_delay = (
            CONF.api_settings.udp_connect_min_interval_health_monitor)
        if hm_is_type_udp and request.delay < conf_min_delay:
            raise exceptions.ValidationException(detail=_(
                "The request delay value %(delay)s should be larger than "
                "%(conf_min_delay)s for %(type)s health monitor type.") % {
                    'delay': request.delay,
                    'conf_min_delay': conf_min_delay,
                    'type': consts.HEALTH_MONITOR_UDP_CONNECT})

    @wsme_pecan.wsexpose(hm_types.HealthMonitorRootResponse,
                         body=hm_types.HealthMonitorRootPOST, status_code=201)
    def post(self, health_monitor_):
        """Creates a health monitor on a pool."""
        context = pecan_request.context.get('octavia_context')
        health_monitor = health_monitor_.healthmonitor

        pool = self._get_db_pool(context.session, health_monitor.pool_id)

        health_monitor.project_id, provider = self._get_lb_project_id_provider(
            context.session, pool.load_balancer_id)

        self._auth_validate_action(context, health_monitor.project_id,
                                   consts.RBAC_POST)

        if (not CONF.api_settings.allow_ping_health_monitors and
                health_monitor.type == consts.HEALTH_MONITOR_PING):
            raise exceptions.DisabledOption(
                option='type', value=consts.HEALTH_MONITOR_PING)

        if pool.protocol in (lib_consts.PROTOCOL_UDP,
                             lib_consts.PROTOCOL_SCTP):
            self._validate_healthmonitor_request_for_udp_sctp(health_monitor,
                                                              pool.protocol)
        else:
            if health_monitor.type in (consts.HEALTH_MONITOR_UDP_CONNECT,
                                       lib_consts.HEALTH_MONITOR_SCTP):
                raise exceptions.ValidationException(detail=_(
                    "The %(type)s type is only supported for pools of type "
                    "%(protocols)s.") % {
                        'type': health_monitor.type,
                        'protocols': '/'.join((consts.PROTOCOL_UDP,
                                               lib_consts.PROTOCOL_SCTP))})

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        lock_session = db_api.get_session(autocommit=False)
        try:
            if self.repositories.check_quota_met(
                    context.session,
                    lock_session,
                    data_models.HealthMonitor,
                    health_monitor.project_id):
                raise exceptions.QuotaException(
                    resource=data_models.HealthMonitor._name())

            hm_dict = db_prepare.create_health_monitor(
                health_monitor.to_dict(render_unsets=True))

            self._test_lb_and_listener_and_pool_statuses(
                lock_session, health_monitor)
            db_hm = self._validate_create_hm(lock_session, hm_dict)

            # Prepare the data for the driver data model
            provider_healthmon = (driver_utils.db_HM_to_provider_HM(db_hm))

            # Dispatch to the driver
            LOG.info("Sending create Health Monitor %s to provider %s",
                     db_hm.id, driver.name)
            driver_utils.call_provider(
                driver.name, driver.health_monitor_create, provider_healthmon)

            lock_session.commit()
        except odb_exceptions.DBError as e:
            lock_session.rollback()
            raise exceptions.InvalidOption(
                value=hm_dict.get('type'), option='type') from e
        except Exception:
            with excutils.save_and_reraise_exception():
                lock_session.rollback()

        db_hm = self._get_db_hm(context.session, db_hm.id)
        result = self._convert_db_to_type(
            db_hm, hm_types.HealthMonitorResponse)
        return hm_types.HealthMonitorRootResponse(healthmonitor=result)

    def _graph_create(self, lock_session, hm_dict):
        hm_dict = db_prepare.create_health_monitor(hm_dict)
        db_hm = self._validate_create_hm(lock_session, hm_dict)

        return db_hm

    def _validate_update_hm(self, db_hm, health_monitor):
        if db_hm.type not in (consts.HEALTH_MONITOR_HTTP,
                              consts.HEALTH_MONITOR_HTTPS):
            if health_monitor.http_method != wtypes.Unset:
                raise exceptions.InvalidOption(
                    value=consts.HTTP_METHOD, option='health monitors of '
                    'type {}'.format(db_hm.type))
            if health_monitor.url_path != wtypes.Unset:
                raise exceptions.InvalidOption(
                    value=consts.URL_PATH, option='health monitors of '
                    'type {}'.format(db_hm.type))
            if health_monitor.expected_codes != wtypes.Unset:
                raise exceptions.InvalidOption(
                    value=consts.EXPECTED_CODES, option='health monitors of '
                    'type {}'.format(db_hm.type))
        if health_monitor.delay is None:
            raise exceptions.InvalidOption(value=None, option=consts.DELAY)
        if health_monitor.max_retries is None:
            raise exceptions.InvalidOption(value=None,
                                           option=consts.MAX_RETRIES)
        if health_monitor.timeout is None:
            raise exceptions.InvalidOption(value=None, option=consts.TIMEOUT)

        if health_monitor.domain_name and not (
                db_hm.http_version or health_monitor.http_version):
            raise exceptions.ValidationException(
                detail=_("'http_version' must be specified when 'domain_name' "
                         "is provided."))

        if ((db_hm.http_version or health_monitor.http_version) and
                (db_hm.domain_name or health_monitor.domain_name)):
            http_version = health_monitor.http_version or db_hm.http_version
            if http_version < 1.1:
                raise exceptions.InvalidOption(
                    value='http_version %s' % http_version,
                    option='health monitors HTTP 1.1 domain name health check')

    def _set_default_on_none(self, health_monitor):
        """Reset settings to their default values if None/null was passed in

        A None/null value can be passed in to clear a value. PUT values
        that were not provided by the user have a type of wtypes.UnsetType.
        If the user is attempting to clear values, they should either
        be set to None (for example in the name field) or they should be
        reset to their default values.
        This method is intended to handle those values that need to be set
        back to a default value.
        """
        if health_monitor.http_method is None:
            health_monitor.http_method = (
                consts.HEALTH_MONITOR_HTTP_DEFAULT_METHOD)
        if health_monitor.url_path is None:
            health_monitor.url_path = (
                consts.HEALTH_MONITOR_DEFAULT_URL_PATH)
        if health_monitor.expected_codes is None:
            health_monitor.expected_codes = (
                consts.HEALTH_MONITOR_DEFAULT_EXPECTED_CODES)
        if health_monitor.max_retries_down is None:
            health_monitor.max_retries_down = consts.DEFAULT_MAX_RETRIES_DOWN

    @wsme_pecan.wsexpose(hm_types.HealthMonitorRootResponse, wtypes.text,
                         body=hm_types.HealthMonitorRootPUT, status_code=200)
    def put(self, id, health_monitor_):
        """Updates a health monitor."""
        context = pecan_request.context.get('octavia_context')
        health_monitor = health_monitor_.healthmonitor
        db_hm = self._get_db_hm(context.session, id, show_deleted=False)

        pool = self._get_db_pool(context.session, db_hm.pool_id)
        project_id, provider = self._get_lb_project_id_provider(
            context.session, pool.load_balancer_id)

        self._auth_validate_action(context, project_id, consts.RBAC_PUT)

        self._validate_update_hm(db_hm, health_monitor)
        # Validate health monitor update options for UDP/SCTP
        if pool.protocol in (lib_consts.PROTOCOL_UDP,
                             lib_consts.PROTOCOL_SCTP):
            health_monitor.type = db_hm.type
            self._validate_healthmonitor_request_for_udp_sctp(health_monitor,
                                                              pool.protocol)

        self._set_default_on_none(health_monitor)

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        with db_api.get_lock_session() as lock_session:

            self._test_lb_and_listener_and_pool_statuses(lock_session, db_hm)

            # Prepare the data for the driver data model
            healthmon_dict = health_monitor.to_dict(render_unsets=False)
            healthmon_dict['id'] = id
            provider_healthmon_dict = (
                driver_utils.hm_dict_to_provider_dict(healthmon_dict))

            # Also prepare the baseline object data
            old_provider_healthmon = driver_utils.db_HM_to_provider_HM(db_hm)

            # Dispatch to the driver
            LOG.info("Sending update Health Monitor %s to provider %s",
                     id, driver.name)
            driver_utils.call_provider(
                driver.name, driver.health_monitor_update,
                old_provider_healthmon,
                driver_dm.HealthMonitor.from_dict(provider_healthmon_dict))

            # Update the database to reflect what the driver just accepted
            health_monitor.provisioning_status = consts.PENDING_UPDATE
            db_hm_dict = health_monitor.to_dict(render_unsets=False)
            self.repositories.health_monitor.update(lock_session, id,
                                                    **db_hm_dict)

        # Force SQL alchemy to query the DB, otherwise we get inconsistent
        # results
        context.session.expire_all()
        db_hm = self._get_db_hm(context.session, id)
        result = self._convert_db_to_type(
            db_hm, hm_types.HealthMonitorResponse)
        return hm_types.HealthMonitorRootResponse(healthmonitor=result)

    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, id):
        """Deletes a health monitor."""
        context = pecan_request.context.get('octavia_context')
        db_hm = self._get_db_hm(context.session, id, show_deleted=False)

        pool = self._get_db_pool(context.session, db_hm.pool_id)
        project_id, provider = self._get_lb_project_id_provider(
            context.session, pool.load_balancer_id)

        self._auth_validate_action(context, project_id, consts.RBAC_DELETE)

        if db_hm.provisioning_status == consts.DELETED:
            return

        # Load the driver early as it also provides validation
        driver = driver_factory.get_driver(provider)

        with db_api.get_lock_session() as lock_session:

            self._test_lb_and_listener_and_pool_statuses(lock_session, db_hm)

            self.repositories.health_monitor.update(
                lock_session, db_hm.id,
                provisioning_status=consts.PENDING_DELETE)

            LOG.info("Sending delete Health Monitor %s to provider %s",
                     id, driver.name)
            provider_healthmon = driver_utils.db_HM_to_provider_HM(db_hm)
            driver_utils.call_provider(
                driver.name, driver.health_monitor_delete, provider_healthmon)



# Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils

from octavia.controller.worker import amphora_rate_limit
import octavia.tests.unit.base as base

AMP_ID = uuidutils.generate_uuid()
BUILD_PRIORITY = 40
USED_BUILD_SLOTS = 0


class TestAmphoraBuildRateLimit(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))

        self.rate_limit = amphora_rate_limit.AmphoraBuildRateLimit()
        self.amp_build_slots_repo = mock.MagicMock()
        self.amp_build_req_repo = mock.MagicMock()
        self.conf.config(group='haproxy_amphora', build_rate_limit=1)

    @mock.patch('octavia.db.api.get_session', mock.MagicMock())
    @mock.patch('octavia.controller.worker.amphora_rate_limit'
                '.AmphoraBuildRateLimit.wait_for_build_slot')
    @mock.patch('octavia.db.repositories.AmphoraBuildReqRepository'
                '.add_to_build_queue')
    def test_add_to_build_request_queue(self,
                                        mock_add_to_build_queue,
                                        mock_wait_for_build_slot):
        self.rate_limit.add_to_build_request_queue(AMP_ID, BUILD_PRIORITY)

        mock_add_to_build_queue.assert_called_once()
        mock_wait_for_build_slot.assert_called_once()

    @mock.patch('octavia.db.api.get_session', mock.MagicMock())
    @mock.patch('octavia.db.repositories.AmphoraBuildSlotsRepository'
                '.get_used_build_slots_count',
                return_value=USED_BUILD_SLOTS)
    def test_has_build_slot(self, mock_get_used_build_slots_count):
        result = self.rate_limit.has_build_slot()

        mock_get_used_build_slots_count.assert_called_once()
        self.assertTrue(result)

    @mock.patch('octavia.db.api.get_session', mock.MagicMock())
    @mock.patch('octavia.db.repositories.AmphoraBuildReqRepository'
                '.get_highest_priority_build_req', return_value=AMP_ID)
    def test_has_highest_priority(self, mock_get_highest_priority_build_req):
        result = self.rate_limit.has_highest_priority(AMP_ID)

        mock_get_highest_priority_build_req.assert_called_once()
        self.assertTrue(result)

    @mock.patch('octavia.db.api.get_session', mock.MagicMock())
    @mock.patch('octavia.db.repositories.AmphoraBuildReqRepository'
                '.update_req_status')
    @mock.patch('octavia.db.repositories.AmphoraBuildSlotsRepository'
                '.update_count')
    def test_update_build_status_and_available_build_slots(self,
                                                           mock_update_count,
                                                           mock_update_status):
        self.rate_limit.update_build_status_and_available_build_slots(AMP_ID)

        mock_update_count.assert_called_once()
        mock_update_status.assert_called_once()

    @mock.patch('octavia.db.api.get_session', mock.MagicMock())
    @mock.patch('octavia.db.repositories.AmphoraBuildReqRepository.delete')
    @mock.patch('octavia.db.repositories.AmphoraBuildSlotsRepository'
                '.update_count')
    def test_remove_from_build_req_queue(self,
                                         mock_update_count,
                                         mock_delete):
        self.rate_limit.remove_from_build_req_queue(AMP_ID)

        mock_update_count.assert_called_once()
        mock_delete.assert_called_once()

    @mock.patch('octavia.db.api.get_session', mock.MagicMock())
    @mock.patch('octavia.db.repositories.AmphoraBuildReqRepository'
                '.delete_all')
    @mock.patch('octavia.db.repositories.AmphoraBuildSlotsRepository'
                '.update_count')
    def test_remove_all_from_build_req_queue(self,
                                             mock_update_count,
                                             mock_delete_all):
        self.rate_limit.remove_all_from_build_req_queue()

        mock_update_count.assert_called_once()
        mock_delete_all.assert_called_once()

    @mock.patch('octavia.controller.worker.amphora_rate_limit'
                '.AmphoraBuildRateLimit.has_build_slot', return_value=True)
    @mock.patch('octavia.controller.worker.amphora_rate_limit'
                '.AmphoraBuildRateLimit.has_highest_priority',
                return_value=True)
    @mock.patch('octavia.controller.worker.amphora_rate_limit'
                '.AmphoraBuildRateLimit.'
                'update_build_status_and_available_build_slots')
    @mock.patch('octavia.controller.worker.amphora_rate_limit'
                '.AmphoraBuildRateLimit.remove_all_from_build_req_queue')
    @mock.patch('time.sleep')
    def test_wait_for_build_slot(self,
                                 mock_time_sleep,
                                 mock_remove_all,
                                 mock_update_status_and_slots_count,
                                 mock_has_high_priority,
                                 mock_has_build_slot):
        self.rate_limit.wait_for_build_slot(AMP_ID)

        self.assertTrue(mock_has_build_slot.called)
        self.assertTrue(mock_has_high_priority.called)



# Copyright 2011-2014 OpenStack Foundation,author: Min Wang,German Eichberger
# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class AmphoraLoadBalancerDriver(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def update_amphora_listeners(self, loadbalancer, amphora,
                                 timeout_dict):
        """Update the amphora with a new configuration.

        :param loadbalancer: List of listeners to update.
        :type loadbalancer: list(octavia.db.models.Listener)
        :param amphora: The index of the specific amphora to update
        :type amphora: octavia.db.models.Amphora
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :type timeout_dict: dict
        :returns: None

        Builds a new configuration, pushes it to the amphora, and reloads
        the listener on one amphora.
        """

    @abc.abstractmethod
    def update(self, loadbalancer):
        """Update the amphora with a new configuration.

        :param loadbalancer: loadbalancer object, need to use its
                             vip.ip_address property
        :type loadbalancer: octavia.db.models.LoadBalancer
        :returns: None

        At this moment, we just build the basic structure for testing, will
        add more function along with the development.
        """

    @abc.abstractmethod
    def start(self, loadbalancer, amphora, timeout_dict=None):
        """Start the listeners on the amphora.

        :param loadbalancer: loadbalancer object to start listeners
        :type loadbalancer: octavia.db.models.LoadBalancer
        :param amphora: Amphora to start. If None, start on all amphora
        :type amphora: octavia.db.models.Amphora
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :type timeout_dict: dict
        :returns: return a value list (listener, vip, status flag--enable)

        At this moment, we just build the basic structure for testing, will
        add more function along with the development.
        """

    @abc.abstractmethod
    def reload(self, loadbalancer, amphora, timeout_dict=None):
        """Reload the listeners on the amphora.

        :param loadbalancer: loadbalancer object to reload listeners
        :type loadbalancer: octavia.db.models.LoadBalancer
        :param amphora: Amphora to start. If None, reload on all amphora
        :type amphora: octavia.db.models.Amphora
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :type timeout_dict: dict
        :returns: return a value list (listener, vip, status flag--enable)

        At this moment, we just build the basic structure for testing, will
        add more function along with the development.
        """

    @abc.abstractmethod
    def delete(self, listener):
        """Delete the listener on the vip.

        :param listener: listener object,
                         need to use its protocol_port property
        :type listener: octavia.db.models.Listener
        :returns: return a value list (listener, vip, status flag--delete)

        At this moment, we just build the basic structure for testing, will
        add more function along with the development.
        """

    @abc.abstractmethod
    def get_info(self, amphora, raise_retry_exception=False):
        """Returns information about the amphora.

        :param amphora: amphora object, need to use its id property
        :type amphora: octavia.db.models.Amphora
        :param raise_retry_exception: Flag if outside task should be retried
        :type boolean: False by default
        :returns: return a value list (amphora.id, status flag--'info')

        At this moment, we just build the basic structure for testing, will
        add more function along with the development, eventually, we want it
        to return information as:
        {"Rest Interface": "1.0", "Amphorae": "1.0",
        "packages":{"ha proxy":"1.5"}}
        some information might come from querying the amphora
        """

    @abc.abstractmethod
    def get_diagnostics(self, amphora):
        """Return ceilometer ready diagnostic data.

        :param amphora: amphora object, need to use its id property
        :type amphora: octavia.db.models.Amphora
        :returns: return a value list (amphora.id, status flag--'ge
                  t_diagnostics')

        At this moment, we just build the basic structure for testing, will
        add more function along with the development, eventually, we want it
        run some expensive self tests to determine if the amphora and the lbs
        are healthy the idea is that those tests are triggered more infrequent
        than the health gathering.
        """

    @abc.abstractmethod
    def finalize_amphora(self, amphora):
        """Finalize the amphora before any listeners are configured.

        :param amphora: amphora object, need to use its id property
        :type amphora: octavia.db.models.Amphora
        :returns: None

        At this moment, we just build the basic structure for testing, will
        add more function along with the development. This is a hook for
        drivers who need to do additional work before an amphora becomes ready
        to accept listeners. Please keep in mind that amphora might be kept in
        an offline pool after this call.
        """

    def post_vip_plug(self, amphora, load_balancer, amphorae_network_config,
                      vrrp_port=None, vip_subnet=None):
        """Called after network driver has allocated and plugged the VIP

        :param amphora:
        :type amphora: octavia.db.models.Amphora
        :param load_balancer: A load balancer that just had its vip allocated
                              and plugged in the network driver.
        :type load_balancer: octavia.common.data_models.LoadBalancer
        :param amphorae_network_config: A data model containing information
                                        about the subnets and ports that an
                                        amphorae owns.
        :param vrrp_port: VRRP port associated with the load balancer
        :type vrrp_port: octavia.network.data_models.Port

        :param vip_subnet: VIP subnet associated with the load balancer
        :type vip_subnet: octavia.network.data_models.Subnet

        :type vip_network: octavia.network.data_models.AmphoraNetworkConfig
        :returns: None

        This is to do any additional work needed on the amphorae to plug
        the vip, such as bring up interfaces.
        """

    def post_network_plug(self, amphora, port):
        """Called after amphora added to network

        :param amphora: amphora object, needs id and network ip(s)
        :type amphora: octavia.db.models.Amphora
        :param port: contains information of the plugged port
        :type port: octavia.network.data_models.Port

        This method is optional to implement.  After adding an amphora to a
        network, there may be steps necessary on the amphora to allow it to
        access said network.  Ex: creating an interface on an amphora for a
        neutron network to utilize.
        """

    def upload_cert_amp(self, amphora, pem_file):
        """Upload cert info to the amphora.

        :param amphora: amphora object, needs id and network ip(s)
        :type amphora: octavia.db.models.Amphora
        :param pem_file: a certificate file
        :type pem_file: file object

        Upload cert file to amphora for Controller Communication.
        """

    def update_amphora_agent_config(self, amphora, agent_config):
        """Upload and update the amphora agent configuration.

        :param amphora: amphora object, needs id and network ip(s)
        :type amphora: octavia.db.models.Amphora
        :param agent_config: The new amphora agent configuration file.
        :type agent_config: string
        """

    @abc.abstractmethod
    def get_interface_from_ip(self, amphora, ip_address, timeout_dict=None):
        """Get the interface name from an IP address.

        :param amphora: The amphora to query.
        :type amphora: octavia.db.models.Amphora
        :param ip_address: The IP address to lookup. (IPv4 or IPv6)
        :type ip_address: string
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :type timeout_dict: dict
        """


class VRRPDriverMixin(object, metaclass=abc.ABCMeta):
    """Abstract mixin class for VRRP support in loadbalancer amphorae

    Usage: To plug VRRP support in another service driver XYZ, use:
    @plug_mixin(XYZ)
    class XYZ: ...
    """
    @abc.abstractmethod
    def update_vrrp_conf(self, loadbalancer, amphorae_network_config, amphora,
                         timeout_dict=None):
        """Update amphorae of the loadbalancer with a new VRRP configuration

        :param loadbalancer: loadbalancer object
        :param amphorae_network_config: amphorae network configurations
        :param amphora: The amphora object to update.
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        """

    @abc.abstractmethod
    def stop_vrrp_service(self, loadbalancer):
        """Stop the vrrp services running on the loadbalancer's amphorae

        :param loadbalancer: loadbalancer object
        """

    @abc.abstractmethod
    def start_vrrp_service(self, amphora, timeout_dict=None):
        """Start the VRRP services on the amphora

        :param amphora: The amphora object to start the service on.
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        """

    @abc.abstractmethod
    def reload_vrrp_service(self, loadbalancer):
        """Reload the VRRP services of all amphorae of the loadbalancer

        :param loadbalancer: loadbalancer object
        """



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from cinderclient import exceptions as cinder_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from tenacity import retry
from tenacity import stop_after_attempt

from octavia.common import clients
from octavia.common import constants
from octavia.common import exceptions
from octavia.volume import volume_base

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class VolumeManager(volume_base.VolumeBase):
    '''Volume implementation of virtual machines via cinder.'''

    def __init__(self):
        super().__init__()
        # Must initialize cinder api
        self._cinder_client = clients.CinderAuth.get_cinder_client(
            service_name=CONF.cinder.service_name,
            endpoint=CONF.cinder.endpoint,
            region=CONF.cinder.region_name,
            endpoint_type=CONF.cinder.endpoint_type,
            insecure=CONF.cinder.insecure,
            cacert=CONF.cinder.ca_certificates_file
        )
        self.manager = self._cinder_client.volumes

    @retry(reraise=True,
           stop=stop_after_attempt(CONF.cinder.volume_create_max_retries))
    def create_volume_from_image(self, image_id):
        """Create cinder volume

        :param image_id: ID of amphora image

        :return volume id
        """
        volume = self.manager.create(
            size=CONF.cinder.volume_size,
            volume_type=CONF.cinder.volume_type,
            availability_zone=CONF.cinder.availability_zone,
            imageRef=image_id)
        resource_status = self.manager.get(volume.id).status

        status = constants.CINDER_STATUS_AVAILABLE
        start = int(time.time())

        while resource_status != status:
            time.sleep(CONF.cinder.volume_create_retry_interval)
            instance_volume = self.manager.get(volume.id)
            resource_status = instance_volume.status
            if resource_status == constants.CINDER_STATUS_ERROR:
                LOG.error('Error creating %s', instance_volume.id)
                instance_volume.delete()
                raise cinder_exceptions.ResourceInErrorState(
                    obj=volume, fault_msg='Cannot create volume')
            if int(time.time()) - start >= CONF.cinder.volume_create_timeout:
                LOG.error('Timed out waiting to create cinder volume %s',
                          instance_volume.id)
                instance_volume.delete()
                raise cinder_exceptions.TimeoutException(
                    obj=volume, action=constants.CINDER_ACTION_CREATE_VOLUME)
        return volume.id

    def delete_volume(self, volume_id):
        """Get glance image from volume

        :param volume_id: ID of amphora boot volume

        :return image id
        """
        LOG.debug('Deleting cinder volume %s', volume_id)
        try:
            instance_volume = self.manager.get(volume_id)
            try:
                instance_volume.delete()
                LOG.debug("Deleted volume %s", volume_id)
            except Exception as e:
                LOG.exception("Error deleting cinder volume %s",
                              volume_id)
                raise exceptions.VolumeDeleteException() from e
        except cinder_exceptions.NotFound:
            LOG.warning("Volume %s not found: assuming already deleted",
                        volume_id)

    def get_image_from_volume(self, volume_id):
        """Get glance image from volume

        :param volume_id: ID of amphora boot volume

        :return image id
        """
        image_id = None
        LOG.debug('Get glance image for volume %s', volume_id)
        try:
            instance_volume = self.manager.get(volume_id)
        except cinder_exceptions.NotFound as e:
            LOG.exception("Volume %s not found", volume_id)
            raise exceptions.VolumeGetException() from e
        if hasattr(instance_volume, 'volume_image_metadata'):
            image_id = instance_volume.volume_image_metadata.get("image_id")
        else:
            LOG.error("Volume %s has no image metadata", volume_id)
            image_id = None
        return image_id



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ssl
from unittest import mock

from octavia.cmd import agent
from octavia.tests.unit import base


class TestAmphoraAgentCMD(base.TestCase):

    def setUp(self):
        super().setUp()

    @mock.patch('octavia.cmd.agent.AmphoraAgent')
    @mock.patch('octavia.amphorae.backends.agent.api_server.server.Server')
    @mock.patch('multiprocessing.Process')
    @mock.patch('octavia.common.service.prepare_service')
    def test_main(self, mock_service, mock_process, mock_server, mock_amp):
        mock_health_proc = mock.MagicMock()
        mock_server_instance = mock.MagicMock()
        mock_amp_instance = mock.MagicMock()

        mock_process.return_value = mock_health_proc
        mock_server.return_value = mock_server_instance
        mock_amp.return_value = mock_amp_instance

        agent.main()

        # Ensure gunicorn is initialized with the correct cert_reqs option.
        # This option is what enforces use of a valid client certificate.
        self.assertEqual(
            ssl.CERT_REQUIRED,
            mock_amp.call_args[0][1]['cert_reqs'])

        mock_health_proc.start.assert_called_once_with()
        mock_amp_instance.run.assert_called_once()



# Copyright 2022 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import signal
from unittest import mock

from octavia.cmd import prometheus_proxy
from octavia.tests.unit import base


class TestPrometheusProxyCMD(base.TestCase):

    @mock.patch('http.server.SimpleHTTPRequestHandler.log_request')
    @mock.patch('http.server.SimpleHTTPRequestHandler.__init__')
    def test_log_request(self, mock_req_handler_init, mock_log_request):
        mock_req_handler_init.return_value = None
        proxy = prometheus_proxy.PrometheusProxy()
        proxy.log_request()
        mock_log_request.assert_not_called()

    @mock.patch('os.cpu_count', return_value=2)
    @mock.patch('psutil.getloadavg', return_value=(1, 2, 3))
    @mock.patch('http.server.SimpleHTTPRequestHandler.__init__')
    def test_add_cpu_utilization(self, mock_req_handler_init, mock_getloadavg,
                                 mock_cpu_count):
        mock_req_handler_init.return_value = None
        proxy = prometheus_proxy.PrometheusProxy()
        test_buffer = "TestStringBuffer\n"
        result = proxy._add_cpu_utilization(test_buffer)

        expected_result = (
            "TestStringBuffer\n"
            "# HELP octavia_loadbalancer_cpu Load balancer CPU utilization "
            "(percentage).\n"
            "# TYPE octavia_loadbalancer_cpu gauge\n"
            "octavia_loadbalancer_cpu 50.0\n")

        self.assertEqual(expected_result, result)

    @mock.patch('psutil.virtual_memory', return_value=(1, 2, 23.5))
    @mock.patch('http.server.SimpleHTTPRequestHandler.__init__')
    def test__add_memory_utilization(self, mock_req_handler_init,
                                     mock_virt_mem):
        mock_req_handler_init.return_value = None
        proxy = prometheus_proxy.PrometheusProxy()
        test_buffer = "TestStringMemoryBuffer\n"
        result = proxy._add_memory_utilization(test_buffer)

        expected_result = (
            "TestStringMemoryBuffer\n"
            "# HELP octavia_loadbalancer_memory Load balancer memory "
            "utilization (percentage).\n"
            "# TYPE octavia_loadbalancer_memory gauge\n"
            "octavia_loadbalancer_memory 23.5\n")

        self.assertEqual(expected_result, result)

    @mock.patch('octavia.cmd.prometheus_proxy.PRINT_REJECTED', True)
    # No need to print all of the rejected lines to the log
    @mock.patch('builtins.print')
    @mock.patch('urllib.request.urlopen')
    @mock.patch('os.cpu_count', return_value=2)
    @mock.patch('psutil.getloadavg', return_value=(1, 2, 3))
    @mock.patch('psutil.virtual_memory', return_value=(1, 2, 23.5))
    @mock.patch('http.server.SimpleHTTPRequestHandler.__init__')
    def test_do_get(self, mock_req_handler_init, mock_virt_mem,
                    mock_getloadavg, mock_cpu_count, mock_urlopen, mock_print):
        mock_req_handler_init.return_value = None
        proxy = prometheus_proxy.PrometheusProxy()

        mock_send_response = mock.MagicMock()
        proxy.send_response = mock_send_response
        mock_send_header = mock.MagicMock()
        proxy.send_header = mock_send_header
        mock_end_headers = mock.MagicMock()
        proxy.end_headers = mock_end_headers
        mock_wfile = mock.MagicMock()
        proxy.wfile = mock_wfile

        with open("octavia/tests/common/sample_haproxy_prometheus",
                  "rb") as file:
            mock_urlopen.return_value = file

            proxy.do_GET()

            mock_send_response.assert_called_once_with(200)

            with open("octavia/tests/common/sample_octavia_prometheus",
                      "rb") as file2:
                octavia_metrics = file2.read()
                mock_wfile.write.assert_called_once_with(octavia_metrics)

    @mock.patch('urllib.request.urlopen')
    @mock.patch('os.cpu_count', return_value=2)
    @mock.patch('psutil.getloadavg', return_value=(1, 2, 3))
    @mock.patch('psutil.virtual_memory', return_value=(1, 2, 23.5))
    @mock.patch('http.server.SimpleHTTPRequestHandler.__init__')
    def test_do_get_exception(self, mock_req_handler_init, mock_virt_mem,
                              mock_getloadavg, mock_cpu_count, mock_urlopen):
        mock_urlopen.side_effect = [Exception('boom')]
        mock_req_handler_init.return_value = None
        proxy = prometheus_proxy.PrometheusProxy()

        mock_send_response = mock.MagicMock()
        proxy.send_response = mock_send_response
        mock_send_header = mock.MagicMock()
        proxy.send_header = mock_send_header
        mock_end_headers = mock.MagicMock()
        proxy.end_headers = mock_end_headers

        proxy.do_GET()

        mock_send_response.assert_called_once_with(502)

    @mock.patch('signal.signal')
    def test_signalhandler(self, mock_signal):

        sig_handler = prometheus_proxy.SignalHandler()

        calls = [mock.call(signal.SIGINT, sig_handler.shutdown),
                 mock.call(signal.SIGTERM, sig_handler.shutdown)]
        mock_signal.assert_has_calls(calls)

        self.assertFalse(prometheus_proxy.EXIT_EVENT.is_set())
        sig_handler.shutdown()
        self.assertTrue(prometheus_proxy.EXIT_EVENT.is_set())

    @mock.patch('octavia.cmd.prometheus_proxy.EXIT_EVENT')
    @mock.patch('signal.signal')
    def test_shutdown_thread(self, mock_signal, mock_exit_event):

        mock_http = mock.MagicMock()

        prometheus_proxy.shutdown_thread(mock_http)

        mock_exit_event.wait.assert_called_once()
        mock_http.shutdown.assert_called_once()

    @mock.patch('threading.Thread')
    # TODO(johnsom) Switch this when we move to ThreadingHTTPServer
    # @mock.patch('http.server.ThreadingHTTPServer.serve_forever')
    @mock.patch('octavia.cmd.prometheus_proxy.ThreadedHTTPServer.'
                'serve_forever')
    @mock.patch('octavia.amphorae.backends.utils.network_namespace.'
                'NetworkNamespace.__exit__')
    @mock.patch('octavia.amphorae.backends.utils.network_namespace.'
                'NetworkNamespace.__enter__')
    @mock.patch('octavia.cmd.prometheus_proxy.EXIT_EVENT')
    @mock.patch('octavia.cmd.prometheus_proxy.SignalHandler')
    def test_main(self, mock_signal_handler, mock_exit_event, mock_netns_enter,
                  mock_netns_exit, mock_serve_forever, mock_thread):

        mock_exit_event.is_set.side_effect = [False, False, True]
        mock_netns_enter.side_effect = [Exception('boom'), True]

        prometheus_proxy.main()

        mock_signal_handler.assert_called_once()
        mock_serve_forever.assert_called_once()



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""rename amphora host id

Revision ID: 3a1e1cdb7b27
Revises: 4faaa983e7a9
Create Date: 2015-01-10 02:01:04.997336

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3a1e1cdb7b27'
down_revision = '4faaa983e7a9'


def upgrade():
    op.alter_column(u'amphora', u'host_id', new_column_name='compute_id',
                    existing_type=sa.String(36), nullable=True)



# Copyright 2018 GoDaddy
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from octavia.statistics import stats_base

LOG = logging.getLogger(__name__)


class StatsLogger(stats_base.StatsDriverMixin):
    def update_stats(self, listener_stats, deltas=False):
        for stats_object in listener_stats:
            LOG.info("Logging listener stats%s for listener `%s` / "
                     "amphora `%s`: %s",
                     ' deltas' if deltas else '',
                     stats_object.listener_id, stats_object.amphora_id,
                     stats_object.get_stats())



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#    under the License.

import os
import stat

import flask
from oslo_config import cfg
import webob

BUFFER = 1024

CONF = cfg.CONF


def upload_server_cert():
    stream = flask.request.stream
    file_path = CONF.amphora_agent.agent_server_cert
    flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    # mode 00600
    mode = stat.S_IRUSR | stat.S_IWUSR
    with os.fdopen(os.open(file_path, flags, mode), 'wb') as crt_file:
        b = stream.read(BUFFER)
        while b:
            crt_file.write(b)
            b = stream.read(BUFFER)

    return webob.Response(json={'message': 'OK'}, status=202)



# Copyright 2018 GoDaddy
# Copyright (c) 2015 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
from unittest import mock

from oslo_utils import uuidutils

from octavia.common import data_models
from octavia.statistics.drivers import update_db
from octavia.tests.unit import base


class TestStatsUpdateDb(base.TestCase):
    def setUp(self):
        super(TestStatsUpdateDb, self).setUp()
        self.amphora_id = uuidutils.generate_uuid()
        self.listener_id = uuidutils.generate_uuid()

    @mock.patch('octavia.db.repositories.ListenerStatisticsRepository')
    @mock.patch('octavia.db.api.get_session')
    def test_update_stats(self, mock_get_session, mock_listener_stats_repo):
        bytes_in1 = random.randrange(1000000000)
        bytes_out1 = random.randrange(1000000000)
        active_conns1 = random.randrange(1000000000)
        total_conns1 = random.randrange(1000000000)
        request_errors1 = random.randrange(1000000000)
        stats_1 = data_models.ListenerStatistics(
            listener_id=self.listener_id,
            amphora_id=self.amphora_id,
            bytes_in=bytes_in1,
            bytes_out=bytes_out1,
            active_connections=active_conns1,
            total_connections=total_conns1,
            request_errors=request_errors1
        )
        bytes_in2 = random.randrange(1000000000)
        bytes_out2 = random.randrange(1000000000)
        active_conns2 = random.randrange(1000000000)
        total_conns2 = random.randrange(1000000000)
        request_errors2 = random.randrange(1000000000)
        stats_2 = data_models.ListenerStatistics(
            listener_id=self.listener_id,
            amphora_id=self.amphora_id,
            bytes_in=bytes_in2,
            bytes_out=bytes_out2,
            active_connections=active_conns2,
            total_connections=total_conns2,
            request_errors=request_errors2
        )

        update_db.StatsUpdateDb().update_stats(
            [stats_1, stats_2], deltas=False)

        mock_listener_stats_repo().replace.assert_has_calls([
            mock.call(mock_get_session(), stats_1),
            mock.call(mock_get_session(), stats_2)
        ])

        update_db.StatsUpdateDb().update_stats(
            [stats_1, stats_2], deltas=True)

        mock_listener_stats_repo().increment.assert_has_calls([
            mock.call(mock_get_session(), stats_1),
            mock.call(mock_get_session(), stats_2)
        ])



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_utils import uuidutils

from octavia.common import exceptions
import octavia.image.drivers.glance_driver as glance_common
import octavia.tests.unit.base as base


class TestGlanceClient(base.TestCase):

    def setUp(self):
        self.manager = glance_common.ImageManager()
        self.manager.manager = mock.MagicMock()

        super(TestGlanceClient, self).setUp()

    def test_no_images(self):
        self.manager.manager.list.return_value = []
        self.assertRaises(
            exceptions.ImageGetException,
            self.manager.get_image_id_by_tag, 'faketag')

    def test_single_image(self):
        images = [
            {'id': uuidutils.generate_uuid(), 'tag': 'faketag'}
        ]
        self.manager.manager.list.return_value = images
        image_id = self.manager.get_image_id_by_tag('faketag', None)
        self.assertEqual(image_id, images[0]['id'])

    def test_single_image_owner(self):
        owner = uuidutils.generate_uuid()
        images = [
            {'id': uuidutils.generate_uuid(),
             'tag': 'faketag',
             'owner': owner}
        ]
        self.manager.manager.list.return_value = images
        image_id = self.manager.get_image_id_by_tag('faketag', owner)
        self.assertEqual(image_id, images[0]['id'])
        self.assertEqual(owner, images[0]['owner'])

    def test_multiple_images_returns_one_of_images(self):
        images = [
            {'id': image_id, 'tag': 'faketag'}
            for image_id in [uuidutils.generate_uuid() for i in range(10)]
        ]
        self.manager.manager.list.return_value = images
        image_id = self.manager.get_image_id_by_tag('faketag', None)
        self.assertIn(image_id, [image['id'] for image in images])



#    Copyright 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""create quotas table

Revision ID: 3f8ff3be828e
Revises: 44a2414dd683
Create Date: 2016-09-01 13:59:20.723621

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3f8ff3be828e'
down_revision = '44a2414dd683'


def upgrade():
    op.create_table(
        u'quotas',
        sa.Column(u'project_id', sa.String(36), primary_key=True,
                  nullable=False),
        sa.Column(u'health_monitor', sa.Integer(), nullable=True),
        sa.Column(u'load_balancer', sa.Integer(), nullable=True),
        sa.Column(u'listener', sa.Integer(), nullable=True),
        sa.Column(u'member', sa.Integer(), nullable=True),
        sa.Column(u'pool', sa.Integer(), nullable=True),
        sa.Column(u'in_use_health_monitor', sa.Integer(), nullable=True),
        sa.Column(u'in_use_load_balancer', sa.Integer(), nullable=True),
        sa.Column(u'in_use_listener', sa.Integer(), nullable=True),
        sa.Column(u'in_use_member', sa.Integer(), nullable=True),
        sa.Column(u'in_use_pool', sa.Integer(), nullable=True),
    )



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import fcntl
import socket
import struct

from oslo_log import log as logging

from octavia.amphorae.backends.utils import network_namespace
from octavia.common import constants
from octavia.common import utils as common_utils

LOG = logging.getLogger(__name__)


def garp(interface, ip_address, net_ns=None):
    """Sends a gratuitous ARP for ip_address on the interface.

    :param interface: The interface name to send the GARP on.
    :param ip_address: The IP address to advertise in the GARP.
    :param net_ns: The network namespace to send the GARP from.
    :returns: None
    """
    ARP_ETHERTYPE = 0x0806
    BROADCAST_MAC = b'\xff\xff\xff\xff\xff\xff'

    # Get a socket, optionally inside a network namespace
    garp_socket = None
    if net_ns:
        with network_namespace.NetworkNamespace(net_ns):
            garp_socket = socket.socket(socket.AF_PACKET, socket.SOCK_RAW)
    else:
        garp_socket = socket.socket(socket.AF_PACKET, socket.SOCK_RAW)

    # Bind the socket with the ARP ethertype protocol
    garp_socket.bind((interface, ARP_ETHERTYPE))

    # Get the MAC address of the interface
    source_mac = garp_socket.getsockname()[4]

    garp_msg = [
        struct.pack('!h', 1),                 # Hardware type ethernet
        struct.pack('!h', 0x0800),            # Protocol type IPv4
        struct.pack('!B', 6),                 # Hardware size
        struct.pack('!B', 4),                 # Protocol size
        struct.pack('!h', 1),                 # Opcode request
        source_mac,                    # Sender MAC address
        socket.inet_aton(ip_address),  # Sender IP address
        BROADCAST_MAC,                 # Target MAC address
        socket.inet_aton(ip_address)]  # Target IP address

    garp_ethernet = [
        BROADCAST_MAC,              # Ethernet destination
        source_mac,                 # Ethernet source
        struct.pack('!h', ARP_ETHERTYPE),  # Ethernet type
        b''.join(garp_msg)]         # The GARP message

    garp_socket.send(b''.join(garp_ethernet))
    garp_socket.close()


def calculate_icmpv6_checksum(packet):
    """Calculate the ICMPv6 checksum for a packet.

    :param packet: The packet bytes to checksum.
    :returns: The checksum integer.
    """
    total = 0

    # Add up 16-bit words
    num_words = len(packet) // 2
    for chunk in struct.unpack("!%sH" % num_words, packet[0:num_words * 2]):
        total += chunk

    # Add any left over byte
    if len(packet) % 2:
        total += packet[-1] << 8

    # Fold 32-bits into 16-bits
    total = (total >> 16) + (total & 0xffff)
    total += total >> 16
    return ~total + 0x10000 & 0xffff


def neighbor_advertisement(interface, ip_address, net_ns=None):
    """Sends a unsolicited neighbor advertisement for an ip on the interface.

    :param interface: The interface name to send the GARP on.
    :param ip_address: The IP address to advertise in the GARP.
    :param net_ns: The network namespace to send the GARP from.
    :returns: None
    """
    ALL_NODES_ADDR = 'ff02::1'
    SIOCGIFHWADDR = 0x8927

    # Get a socket, optionally inside a network namespace
    na_socket = None
    if net_ns:
        with network_namespace.NetworkNamespace(net_ns):
            na_socket = socket.socket(
                socket.AF_INET6, socket.SOCK_RAW,
                socket.getprotobyname(constants.IPV6_ICMP))
    else:
        na_socket = socket.socket(socket.AF_INET6, socket.SOCK_RAW,
                                  socket.getprotobyname(constants.IPV6_ICMP))

    # Per RFC 4861 section 4.4, the hop limit should be 255
    na_socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, 255)

    # Bind the socket with the source address
    na_socket.bind((ip_address, 0))

    # Get the byte representation of the MAC address of the interface
    # Note: You can't use getsockname() to get the MAC on this type of socket
    source_mac = fcntl.ioctl(
        na_socket.fileno(), SIOCGIFHWADDR, struct.pack(
            '256s', bytes(interface, 'utf-8')))[18:24]

    # Get the byte representation of the source IP address
    source_ip_bytes = socket.inet_pton(socket.AF_INET6, ip_address)

    icmpv6_na_msg_prefix = [
        struct.pack('!B', 136),   # ICMP Type Neighbor Advertisement
        struct.pack('!B', 0)]     # ICMP Code
    icmpv6_na_msg_postfix = [
        struct.pack('!I', 0xa0000000),  # Flags (Router, Override)
        source_ip_bytes,         # Target address
        struct.pack('!B', 2),    # ICMPv6 option type target link-layer address
        struct.pack('!B', 1),    # ICMPv6 option length
        source_mac]              # ICMPv6 option link-layer address

    # Calculate the ICMPv6 checksum
    icmpv6_pseudo_header = [
        source_ip_bytes,         # Source IP address
        socket.inet_pton(socket.AF_INET6, ALL_NODES_ADDR),  # Destination IP
        struct.pack('!I', 58),   # IPv6 next header (ICMPv6)
        struct.pack('!h', 32)]   # IPv6 payload length
    icmpv6_tmp_chksum = struct.pack('!H', 0)  # Checksum->zeros for calculation
    tmp_chksum_msg = b''.join(icmpv6_pseudo_header + icmpv6_na_msg_prefix +
                              [icmpv6_tmp_chksum] + icmpv6_pseudo_header)
    checksum = struct.pack('!H', calculate_icmpv6_checksum(tmp_chksum_msg))

    # Build the ICMPv6 unsolicitated neighbor advertisement
    icmpv6_msg = b''.join(icmpv6_na_msg_prefix + [checksum] +
                          icmpv6_na_msg_postfix)

    na_socket.sendto(icmpv6_msg, (ALL_NODES_ADDR, 0, 0, 0))
    na_socket.close()


def send_ip_advertisement(interface, ip_address, net_ns=None):
    """Send an address advertisement.

    This method will send either GARP (IPv4) or neighbor advertisements (IPv6)
    for the ip address specified.

    :param interface: The interface name to send the advertisement on.
    :param ip_address: The IP address to advertise.
    :param net_ns: The network namespace to send the advertisement from.
    :returns: None
    """
    try:
        if common_utils.is_ipv4(ip_address):
            garp(interface, ip_address, net_ns)
        elif common_utils.is_ipv6(ip_address):
            neighbor_advertisement(interface, ip_address, net_ns)
        else:
            LOG.error('Unknown IP version for address: "%s". Skipping',
                      ip_address)
    except Exception as e:
        LOG.warning('Unable to send address advertisement for address: "%s", '
                    'error: %s. Skipping', ip_address, str(e))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add lb flavor ID constraint

Revision ID: e37941b010db
Revises: dcf88e59aae4
Create Date: 2019-10-31 10:09:37.869653

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

from octavia.common import constants

# revision identifiers, used by Alembic.
revision = 'e37941b010db'
down_revision = 'dcf88e59aae4'


def upgrade():
    insert_table = sql.table(
        u'flavor_profile',
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=False),
        sa.Column(u'provider_name', sa.String(255), nullable=False),
        sa.Column(u'flavor_data', sa.String(4096), nullable=False),
    )

    op.bulk_insert(
        insert_table,
        [
            {'id': constants.NIL_UUID, 'name': 'DELETED-PLACEHOLDER',
             'provider_name': 'DELETED', 'flavor_data': '{}'},
        ]
    )

    insert_table = sql.table(
        u'flavor',
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=False),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.Column(u'flavor_profile_id', sa.String(36), nullable=False),
    )

    op.bulk_insert(
        insert_table,
        [
            {'id': constants.NIL_UUID, 'name': 'DELETED-PLACEHOLDER',
             'description': 'Placeholder for DELETED LBs with DELETED flavors',
             'enabled': False, 'flavor_profile_id': constants.NIL_UUID}
        ]
    )

    # Make sure any existing load balancers with invalid flavor_id
    # map to a valid flavor.
    # Note: constant is not used here to not trigger security tool errors.
    op.execute("UPDATE load_balancer LEFT JOIN flavor ON "
               "load_balancer.flavor_id = flavor.id SET "
               "load_balancer.flavor_id = "
               "'00000000-0000-0000-0000-000000000000' WHERE "
               "flavor.id IS NULL and load_balancer.flavor_id IS NOT NULL")

    op.create_foreign_key('fk_loadbalancer_flavor_id', 'load_balancer',
                          'flavor', ['flavor_id'], ['id'])
